@@ -1,6 +1,6 @@
 /*###################################################################################
 #
-#   Embperl - Copyright (c) 1997-1998 Gerald Richter / ECOS
+#   Embperl - Copyright (c) 1997-1999 Gerald Richter / ECOS
 #
 #   You may distribute under the terms of either the GNU General Public
 #   License or the Artistic License, as specified in the Perl README file.
@@ -60,6 +60,15 @@ enum tRc
     rcEndforeachWithoutForeach, 
     rcMissingArgs,
     rcNotAnArray,
+    rcCallInputFuncFailed,
+    rcCallOutputFuncFailed,
+    rcSubNotFound,
+    rcImportStashErr,
+    rcCGIError,
+    rcUnclosedHtml,
+    rcUnclosedCmd,
+    rcNotAllowed
+    
     } ;
 
 
@@ -89,6 +98,9 @@ enum dbg
     dbgWatchScalar      = 0x20000,
     dbgHeadersIn        = 0x40000,
     dbgShowCleanup      = 0x80000,
+    dbgProfile          = 0x100000,
+    dbgSession          = 0x200000,
+    dbgImport		= 0x400000,
     
     dbgAll  = -1
     } ;
@@ -115,6 +127,13 @@ enum opt
     optAllFormData             = 8192,
     optRedirectStdout          = 16384,
     optUndefToEmptyValue       = 32768,
+    optNoHiddenEmptyValue      = 65536,
+    optAllowZeroFilesize       = 0x20000, 
+    optReturnError             = 0x40000, 
+    optKeepSrcInMemory         = 0x80000,
+    optKeepSpaces	       = 0x100000,
+    optOpenLogEarly            = 0x200000,
+    optNoUncloseWarn	       = 0x400000
     } ;
 
 /*
@@ -145,10 +164,7 @@ enum epIO
 
 
 
-#define ERRDATLEN 1024
 
-extern char errdat1 [ERRDATLEN]  ;
-extern char errdat2 [ERRDATLEN]  ;
 
 /*
     Escape modes
