@@ -1,7 +1,7 @@
-#!/usr/bin/perl5.00404 --
+#!/usr/bin/perl 
 ###################################################################################
 #
-#   Embperl - Copyright (c) 1997-1998 Gerald Richter / ECOS
+#   Embperl - Copyright (c) 1997-1999 Gerald Richter / ECOS
 #
 #   You may distribute under the terms of either the GNU General Public
 #   License or the Artistic License, as specified in the Perl README file.
@@ -13,8 +13,11 @@
 #
 ###################################################################################
 
+
 use HTML::Embperl;
 
+die "Do not use as CGI script. Use 'embpcgi.pl' instead" if ($ENV{PATH_TRANSLATED}) ;
+
 HTML::Embperl::run (@ARGV) ;
 
 
