@@ -12,13 +12,14 @@
 #
 ###################################################################################*/
 
+/* input and output escaping for iso-8859-1 (iso-latin-1) */
 
 #include "ep.h"
 
 
 
 /*
-/* Character Translation
+* Character Translation
 */
 
 
@@ -192,18 +193,18 @@ struct tCharTrans Char2Html [] =
         { '' ,   "&pound;"  },    /* 	Pound sterling  */
         { '' ,   "&curren;" },    /* 	General currency sign  */
         { '' ,   "&yen;"    },    /* 	Yen sign  */
-/*        { '' ,   "&brvbar;" },    /*  	Broken vertical bar  */
+/*        { '' ,   "&brvbar;" },    / *  	Broken vertical bar  */
         { '' ,   "&brkbar;" },    /* 	Broken vertical bar  */
-        { '' ,   "&&sect;"  },    /* 	Section sign  */
-/*        { '' ,   "&&um;"    },    /*  	Diresis / Umlaut  */
-        { '' ,   "&&die;"   },    /* 	Diresis / Umlaut  */
-        { '' ,   "&&copy;"  },    /* 	Copyright               */
-        { '' ,   "&&ordf;"  },    /* 	Feminine ordinal  */
-        { '' ,   "&&laquo;" },    /* 	Left angle quote, guillemot left  */
-        { '' ,   "&&not;"   },    /*	Not sign  */
+        { '' ,   "&sect;"  },    /* 	Section sign  */
+/*        { '' ,   "&&um;"    },    / *  	Diresis / Umlaut  */
+        { '' ,   "&die;"   },    /* 	Diresis / Umlaut  */
+        { '' ,   "&copy;"  },    /* 	Copyright               */
+        { '' ,   "&ordf;"  },    /* 	Feminine ordinal  */
+        { '' ,   "&laquo;" },    /* 	Left angle quote, guillemot left  */
+        { '' ,   "&not;"   },    /*	Not sign  */
         { '' ,   "&shy;"    },    /* 	Soft hyphen  */
         { '' ,   "&reg;"    },    /* 	Registered trademark  */
-/*        { '' ,   "&macr;"   },    /*  	Macron accent  */
+/*        { '' ,   "&macr;"   },    / *  	Macron accent  */
         { '' ,   "&hibar;"  },    /* 	Macron accent  */
         { '' ,   "&deg;"    },    /* 	Degree sign  */
         { '' ,   "&plusmn;" },    /* 	Plus or minus  */
@@ -325,10 +326,10 @@ struct tCharTrans Char2Url [] =
         { ' ' ,   "%1F"         },    /* &#31;		Unused  */
         { ' ' ,   "+"           },    /* 	&#32;		Space  */
         { '!' ,   ""         },    /* 	&#33;		Exclamation mark  */
-        { '"' ,   ""   },    /* 	Quotation mark  */
+        { '"' ,   "%22"   },    /* 	Quotation mark  */
         { '#' ,   ""         },    /* 	&#35;		Number sign  */
         { '$' ,   ""         },    /* 	&#36;		Dollar sign  */
-        { '%' ,   ""         },    /* 	&#37;		Percent sign  */
+        { '%' ,   "%25"      },    /* 	&#37;		Percent sign  */
         { '&' ,   "%26"    },    /* 	Ampersand  */
         { '\'' ,  ""         },    /* 	&#39;		Apostrophe  */
         { '(' ,   ""         },    /* 	&#40;		Left parenthesis  */
@@ -554,13 +555,6 @@ struct tCharTrans Char2Url [] =
 struct tCharTrans Html2Char [] = 
  
     {
-        { '' ,   "&&copy"  },    /* 	Copyright  */
-        { '' ,   "&&die"   },    /* 	Diresis / Umlaut  */
-        { '' ,   "&&laquo" },    /* 	Left angle quote, guillemot left  */
-        { '' ,   "&&not"   },    /*	Not sign  */
-        { '' ,   "&&ordf"  },    /* 	Feminine ordinal  */
-        { '' ,   "&&sect"  },    /* 	Section sign  */
-        { '' ,   "&&um"    },    /*  	Diresis / Umlaut  */
         { '' ,   "&AElig"  },    /* 	Capital AE ligature  */
         { '' ,   "&Aacute" },    /* 	Capital A, acute accent  */
         { '' ,   "&Acirc"  },    /* 	Capital A, circumflex  */
@@ -605,8 +599,10 @@ struct tCharTrans Html2Char [] =
         { '' ,   "&ccedil" },    /* 	Small c, cedilla  */
         { '' ,   "&cedil"  },    /* 	Cedilla  */
         { '' ,   "&cent"   },    /* 	Cent sign  */
+        { '' ,   "&copy"  },    /* 	Copyright  */
         { '' ,   "&curren" },    /* 	General currency sign  */
         { '' ,   "&deg"    },    /* 	Degree sign  */
+        { '' ,   "&die"   },    /* 	Diresis / Umlaut  */
         { '' ,   "&divide" },    /* 	Division sign  */
         { '' ,   "&eacute" },    /* 	Small e, acute accent  */
         { '' ,   "&ecirc"  },    /* 	Small e, circumflex  */
@@ -624,15 +620,18 @@ struct tCharTrans Html2Char [] =
         { '' ,   "&igrave" },    /* 	Small i, grave accent  */
         { '' ,   "&iquest" },    /* 	Inverted question mark  */
         { '' ,   "&iuml"   },    /* 	Small i, diresis / umlaut  */
+        { '' ,   "&laquo" },    /* 	Left angle quote, guillemot left  */
         { '<' ,   "&lt"     },    /* 	Less than  */
         { '' ,   "&macr"   },    /*  	Macron accent  */
         { '' ,   "&micro"  },    /* 	Micro sign  */
         { '' ,   "&middot" },    /* 	Middle dot  */
         { ' ' ,   "&nbsp"   },    /* 	Non-breaking Space  */
+        { '' ,   "&not"   },    /*	Not sign  */
         { '' ,   "&ntilde" },    /* 	Small n, tilde  */
         { '' ,   "&oacute" },    /* 	Small o, acute accent  */
         { '' ,   "&ocirc"  },    /* 	Small o, circumflex  */
         { '' ,   "&ograve" },    /* 	Small o, grave accent  */
+        { '' ,   "&ordf"  },    /* 	Feminine ordinal  */
         { '' ,   "&ordm"   },    /* 	Masculine ordinal  */
         { '' ,   "&oslash" },    /* 	Small o, slash  */
         { '' ,   "&otilde" },    /* 	Small o, tilde  */
@@ -643,6 +642,7 @@ struct tCharTrans Html2Char [] =
         { '"' ,   "&quot"   },    /* 	Quotation mark  */
         { '' ,   "&raquo"  },    /* 	Right angle quote, guillemot right  */
         { '' ,   "&reg"    },    /* 	Registered trademark  */
+        { '' ,   "&sect"  },    /* 	Section sign  */
         { '' ,   "&shy"    },    /* 	Soft hyphen  */
         { '' ,   "&sup1"   },    /* 	Superscript one  */
         { '' ,   "&sup2"   },    /* 	Superscript two  */
@@ -653,6 +653,7 @@ struct tCharTrans Html2Char [] =
         { '' ,   "&uacute" },    /* 	Small u, acute accent  */
         { '' ,   "&ucirc"  },    /* 	Small u, circumflex  */
         { '' ,   "&ugrave" },    /* 	Small u, grave accent  */
+        { '' ,   "&um"    },    /*  	Diresis / Umlaut  */
         { '' ,   "&uuml"   },    /* 	Small u, diresis / umlaut  */
         { '' ,   "&yacute" },    /* 	Small y, acute accent  */
         { '' ,   "&yen"    },    /* 	Yen sign  */
@@ -662,8 +663,3 @@ struct tCharTrans Html2Char [] =
 
 int sizeHtml2Char = sizeof (Html2Char) / sizeof (struct tCharTrans) ;
 
-struct tCharTrans * pCurrEscape = Char2Html ;
-struct tCharTrans * pNextEscape = Char2Html ;
-
-int bEscMode = escUrl | escHtml ;
-int bEscModeSet = 0 ;
