@@ -1,6 +1,6 @@
 /*###################################################################################
 #
-#   Embperl - Copyright (c) 1997-1998 Gerald Richter / ECOS
+#   Embperl - Copyright (c) 1997-1999 Gerald Richter / ECOS
 #
 #   You may distribute under the terms of either the GNU General Public
 #   License or the Artistic License, as specified in the Perl README file.
@@ -17,11 +17,6 @@
 #include "epmacro.h"
 
 
-int numEvals ;
-int numCacheHits ;
-
-int  bStrict ; /* aply use strict in each eval */
-
 /* -------------------------------------------------------------------------------
 *
 * Eval PERL Statements 
@@ -31,7 +26,8 @@ int  bStrict ; /* aply use strict in each eval */
 *
 ------------------------------------------------------------------------------- */
 
-int EvalDirect (/*in*/    SV * pArg) 
+int EvalDirect (/*i/o*/ register req * r,
+			/*in*/    SV * pArg) 
     {
     dSP;
     SV *  pSVErr  ;
@@ -39,6 +35,7 @@ int EvalDirect (/*in*/    SV * pArg)
     EPENTRY (EvalDirect) ;
 
     tainted = 0 ;
+    pCurrReq = r ;
 
     PUSHMARK(sp);
     perl_eval_sv(pArg, G_SCALAR | G_KEEPERR);
@@ -49,14 +46,14 @@ int EvalDirect (/*in*/    SV * pArg)
 	{
         STRLEN l ;
         char * p = SvPV (pSVErr, l) ;
-        if (l > sizeof (errdat1) - 1)
-            l = sizeof (errdat1) - 1 ;
-        strncpy (errdat1, p, l) ;
-        if (l > 0 && errdat1[l-1] == '\n')
+        if (l > sizeof (r -> errdat1) - 1)
+            l = sizeof (r -> errdat1) - 1 ;
+        strncpy (r -> errdat1, p, l) ;
+        if (l > 0 && r -> errdat1[l-1] == '\n')
             l-- ;
-        errdat1[l] = '\0' ;
+        r -> errdat1[l] = '\0' ;
          
-	LogError (rcEvalErr) ;
+	LogError (r, rcEvalErr) ;
 
 	sv_setpv(pSVErr,"");
         return rcEvalErr ;
@@ -76,65 +73,79 @@ int EvalDirect (/*in*/    SV * pArg)
 *
 ------------------------------------------------------------------------------- */
 
-static int EvalAll (/*in*/  const char *  sArg,
+static int EvalAll (/*i/o*/ register req * r,
+		    /*in*/  const char *  sArg,
                     /*in*/  int           flags,
-                    /*out*/ SV **         pRet)             
+                    /*in*/  const char *  sName,
+		    /*out*/ SV **         pRet)             
     {
-    static char sFormat []       = "package %s ; sub { \n#line %d %s\n%s\n}" ;
-    static char sFormatStrict [] = "package %s ; use strict ; sub {\n#line %d %s\n%s\n}" ; 
-    static char sFormatArray []       = "package %s ; sub { \n#line %d %s\n[%s]\n}" ;
-    static char sFormatStrictArray [] = "package %s ; use strict ; sub {\n#line %d %s\n[%s]\n}" ; 
+    static char sFormat []       = "package %s ; sub %s { \n#line %d \"%s\"\n%s\n} %s%s" ;
+    static char sFormatStrict [] = "package %s ; use strict ; sub %s {\n#line %d \"%s\"\n%s\n} %s%s" ; 
+    static char sFormatArray []       = "package %s ; sub %s { \n#line %d \"%s\"\n[%s]\n} %s%s" ;
+    static char sFormatStrictArray [] = "package %s ; use strict ; sub %s {\n#line %d \"%s\"\n[%s]\n} %s%s" ; 
     SV *   pSVCmd ;
     SV *   pSVErr ;
+    int    n ;
+    char * sRef = "" ;
 
     dSP;
     
     EPENTRY (EvalAll) ;
 
-    GetLineNo () ;
+    GetLineNo (r) ;
 
-    if (bDebug & dbgDefEval)
-        lprintf ("[%d]DEF:  Line %d: %s\n", nPid, nSourceline, sArg) ;
+    if (r -> bDebug & dbgDefEval)
+        lprintf (r, "[%d]DEF:  Line %d: %s\n", r -> nPid, r -> Buf.nSourceline, sArg) ;
 
     tainted = 0 ;
+    pCurrReq = r ;
 
-    if (bStrict)
+    if (*sName)
+	sRef = "; \\&" ;
+    
+    if (r -> bStrict)
         if (flags & G_ARRAY)
-            pSVCmd = newSVpvf(sFormatStrictArray, sEvalPackage, nSourceline, sSourcefile, sArg) ;
+            pSVCmd = newSVpvf(sFormatStrictArray, r -> Buf.sEvalPackage, sName, r -> Buf.nSourceline, r -> Buf.pFile -> sSourcefile, sArg, sRef, sName) ;
         else
-            pSVCmd = newSVpvf(sFormatStrict, sEvalPackage, nSourceline, sSourcefile, sArg) ;
+            pSVCmd = newSVpvf(sFormatStrict, r -> Buf.sEvalPackage, sName, r -> Buf.nSourceline, r -> Buf.pFile -> sSourcefile, sArg, sRef, sName) ;
     else
         if (flags & G_ARRAY)
-            pSVCmd = newSVpvf(sFormatArray, sEvalPackage, nSourceline, sSourcefile, sArg) ;
+            pSVCmd = newSVpvf(sFormatArray, r -> Buf.sEvalPackage, sName, r -> Buf.nSourceline, r -> Buf.pFile -> sSourcefile, sArg, sRef, sName) ;
         else
-            pSVCmd = newSVpvf(sFormat, sEvalPackage, nSourceline, sSourcefile, sArg) ;
+            pSVCmd = newSVpvf(sFormat, r -> Buf.sEvalPackage, sName, r -> Buf.nSourceline, r -> Buf.pFile -> sSourcefile, sArg, sRef, sName) ;
 
     PUSHMARK(sp);
-    perl_eval_sv(pSVCmd, G_SCALAR | G_KEEPERR);
+    n = perl_eval_sv(pSVCmd, G_SCALAR | G_KEEPERR);
     SvREFCNT_dec(pSVCmd);
 
     SPAGAIN;
-    *pRet = POPs;
+    if (n > 0)
+        *pRet = POPs;
+    else
+	*pRet = NULL ;
     PUTBACK;
 
-    if (bDebug & dbgMem)
-        lprintf ("[%d]SVs:  %d\n", nPid, sv_count) ;
+    if (r -> bDebug & dbgMem)
+        lprintf (r, "[%d]SVs:  %d\n", r -> nPid, sv_count) ;
     
     pSVErr = ERRSV ;
-    if (SvTRUE (pSVErr))
+    if (SvTRUE (pSVErr) || (n == 0 && (flags & G_DISCARD) == 0))
         {
         STRLEN l ;
         char * p = SvPV (pSVErr, l) ;
-        if (l > sizeof (errdat1) - 1)
-            l = sizeof (errdat1) - 1 ;
-        strncpy (errdat1, p, l) ;
-        if (l > 0 && errdat1[l-1] == '\n')
+        if (l > sizeof (r -> errdat1) - 1)
+            l = sizeof (r -> errdat1) - 1 ;
+        strncpy (r -> errdat1, p, l) ;
+        if (l > 0 && r -> errdat1[l-1] == '\n')
             l-- ;
-        errdat1[l] = '\0' ;
+        r -> errdat1[l] = '\0' ;
          
-	*pRet = newSVpv (errdat1, 0) ;
+        if (pRet && *pRet)
+	     SvREFCNT_dec (*pRet) ;
+	
+	*pRet = newSVpv (r -> errdat1, 0) ;
          
-        LogError (rcEvalErr) ;
+        LogError (r, rcEvalErr) ;
 	sv_setpv(pSVErr, "");
         return rcEvalErr ;
         }
@@ -154,13 +165,14 @@ static int EvalAll (/*in*/  const char *  sArg,
 
 #define EVAL_SUB
 
-static int EvalAllNoCache (/*in*/  const char *  sArg,
+static int EvalAllNoCache (/*i/o*/ register req * r,
+			/*in*/  const char *  sArg,
                            /*out*/ SV **         pRet)             
     {
     int   num ;         
-    int   nCountUsed = TableState.nCountUsed ;
-    int   nRowUsed   = TableState.nRowUsed ;
-    int   nColUsed   = TableState.nColUsed ;
+    int   nCountUsed = r -> TableStack.State.nCountUsed ;
+    int   nRowUsed   = r -> TableStack.State.nRowUsed ;
+    int   nColUsed   = r -> TableStack.State.nColUsed ;
 #ifndef EVAL_SUB    
     SV *  pSVArg ;
 #endif
@@ -169,10 +181,11 @@ static int EvalAllNoCache (/*in*/  const char *  sArg,
 
     EPENTRY (EvalAll) ;
 
-    if (bDebug & dbgEval)
-        lprintf ("[%d]EVAL< %s\n", nPid, sArg) ;
+    if (r -> bDebug & dbgEval)
+        lprintf (r, "[%d]EVAL< %s\n", r -> nPid, sArg) ;
 
     tainted = 0 ;
+    pCurrReq = r ;
 
 #ifdef EVAL_SUB    
 
@@ -186,47 +199,47 @@ static int EvalAllNoCache (/*in*/  const char *  sArg,
     
     pSVArg = sv_2mortal(newSVpv((char *)sArg, strlen (sArg))) ;
 
-    /*num = perl_eval_sv (pSVArg, G_SCALAR) ; /* call the function             */ */
+    /*num = perl_eval_sv (pSVArg, G_SCALAR) ; / * call the function             */ */
     num = perl_eval_sv (pSVArg, G_DISCARD) ; /* call the function             */
     num = 0 ;
 #endif    
     SPAGAIN;                        /* refresh stack pointer         */
     
-    if (bDebug & dbgMem)
-        lprintf ("[%d]SVs:  %d\n", nPid, sv_count) ;
+    if (r -> bDebug & dbgMem)
+        lprintf (r, "[%d]SVs:  %d\n", r -> nPid, sv_count) ;
     /* pop the return value from stack */
     if (num == 1)   
         {
         *pRet = POPs ;
         SvREFCNT_inc (*pRet) ;
 
-        if (bDebug & dbgEval)
+        if (r -> bDebug & dbgEval)
             if (SvOK (*pRet))
-                lprintf ("[%d]EVAL> %s\n", nPid, SvPV (*pRet, na)) ;
+                lprintf (r, "[%d]EVAL> %s\n", r -> nPid, SvPV (*pRet, na)) ;
             else
-                lprintf ("[%d]EVAL> <undefined>\n", nPid) ;
+                lprintf (r, "[%d]EVAL> <undefined>\n", r -> nPid) ;
         
-        if ((nCountUsed != TableState.nCountUsed ||
-             nColUsed != TableState.nColUsed ||
-             nRowUsed != TableState.nRowUsed) &&
+        if ((nCountUsed != r -> TableStack.State.nCountUsed ||
+             nColUsed != r -> TableStack.State.nColUsed ||
+             nRowUsed != r -> TableStack.State.nRowUsed) &&
               !SvOK (*pRet))
             {
-            TableState.nResult = 0 ;
+            r -> TableStack.State.nResult = 0 ;
             SvREFCNT_dec (*pRet) ;
             *pRet = newSVpv("", 0) ;
             } 
 
-        if ((bDebug & dbgTab) &&
-            (TableState.nCountUsed ||
-             TableState.nColUsed ||
-             TableState.nRowUsed))
-            lprintf ("[%d]TAB:  nResult = %d\n", nPid, TableState.nResult) ;
+        if ((r -> bDebug & dbgTab) &&
+            (r -> TableStack.State.nCountUsed ||
+             r -> TableStack.State.nColUsed ||
+             r -> TableStack.State.nRowUsed))
+            lprintf (r, "[%d]TAB:  nResult = %d\n", r -> nPid, r -> TableStack.State.nResult) ;
         }
     else
         {
         *pRet = NULL ;
-        if (bDebug & dbgEval)
-            lprintf ("[%d]EVAL> <NULL>\n", nPid) ;
+        if (r -> bDebug & dbgEval)
+            lprintf (r, "[%d]EVAL> <NULL>\n", r -> nPid) ;
         }
 
 	PUTBACK;
@@ -234,8 +247,8 @@ static int EvalAllNoCache (/*in*/  const char *  sArg,
     pSVErr = ERRSV ;
     if (SvTRUE (pSVErr))
         {
-        strncpy (errdat1, SvPV (pSVErr, na), sizeof (errdat1) - 1) ;
-        LogError (rcEvalErr) ;
+        strncpy (r -> errdat1, SvPV (pSVErr, na), sizeof (r -> errdat1) - 1) ;
+        LogError (r, rcEvalErr) ;
 	num = rcEvalErr ;
         }
     else
@@ -258,7 +271,7 @@ static int EvalAllNoCache (/*in*/  const char *  sArg,
 ------------------------------------------------------------------------------- */
 
 
-static int Watch  ()
+static int Watch (/*i/o*/ register req * r)
 
     {
     dSP;                            /* initialize stack pointer      */
@@ -283,28 +296,27 @@ static int Watch  ()
 ------------------------------------------------------------------------------- */
 
 
-static int CallCV  (/*in*/  const char *  sArg,
+static int CallCV  (/*i/o*/ register req * r,
+		    /*in*/  const char *  sArg,
                     /*in*/  CV *          pSub,
                     /*in*/  int           flags,
                     /*out*/ SV **         pRet)             
     {
     int   num ;         
-    int   nCountUsed = TableState.nCountUsed ;
-    int   nRowUsed   = TableState.nRowUsed ;
-    int   nColUsed   = TableState.nColUsed ;
-    int   bDynTab    = 0 ;
+    int   nCountUsed = r -> TableStack.State.nCountUsed ;
+    int   nRowUsed   = r -> TableStack.State.nRowUsed ;
+    int   nColUsed   = r -> TableStack.State.nColUsed ;
     SV *  pSVErr ;
 
-    SV *  pSVArg ;
     dSP;                            /* initialize stack pointer      */
 
     EPENTRY (CallCV) ;
 
-    if (bDebug & dbgEval)
-        lprintf ("[%d]EVAL< %s\n", nPid, sArg) ;
+    if (r -> bDebug & dbgEval)
+        lprintf (r, "[%d]EVAL< %s\n", r -> nPid, sArg) ;
 
     tainted = 0 ;
-
+    pCurrReq = r ;
 
     ENTER ;
     SAVETMPS ;
@@ -314,8 +326,8 @@ static int CallCV  (/*in*/  const char *  sArg,
     
     SPAGAIN;                        /* refresh stack pointer         */
     
-    if (bDebug & dbgMem)
-        lprintf ("[%d]SVs:  %d\n", nPid, sv_count) ;
+    if (r -> bDebug & dbgMem)
+        lprintf (r, "[%d]SVs:  %d\n", r -> nPid, sv_count) ;
     /* pop the return value from stack */
     if (num == 1)   
         {
@@ -328,52 +340,59 @@ static int CallCV  (/*in*/  const char *  sArg,
         else        
             SvREFCNT_inc (*pRet) ;
 
-        if (bDebug & dbgEval)
+        if (r -> bDebug & dbgEval)
             {
             if (SvOK (*pRet))
-                lprintf ("[%d]EVAL> %s\n", nPid, SvPV (*pRet, na)) ;
+                lprintf (r, "[%d]EVAL> %s\n", r -> nPid, SvPV (*pRet, na)) ;
             else
-                lprintf ("[%d]EVAL> <undefined>\n", nPid) ;
+                lprintf (r, "[%d]EVAL> <undefined>\n", r -> nPid) ;
             }                
             
-        if ((nCountUsed != TableState.nCountUsed ||
-             nColUsed != TableState.nColUsed ||
-             nRowUsed != TableState.nRowUsed) &&
+        if ((nCountUsed != r -> TableStack.State.nCountUsed ||
+             nColUsed != r -> TableStack.State.nColUsed ||
+             nRowUsed != r -> TableStack.State.nRowUsed) &&
               !SvOK (*pRet))
             {
-            TableState.nResult = 0 ;
+            r -> TableStack.State.nResult = 0 ;
             SvREFCNT_dec (*pRet) ;
             *pRet = newSVpv("", 0) ;
             } 
 
-        if ((bDebug & dbgTab) &&
-            (TableState.nCountUsed ||
-             TableState.nColUsed ||
-             TableState.nRowUsed))
-            lprintf ("[%d]TAB:  nResult = %d\n", nPid, TableState.nResult) ;
+        if ((r -> bDebug & dbgTab) &&
+            (r -> TableStack.State.nCountUsed ||
+             r -> TableStack.State.nColUsed ||
+             r -> TableStack.State.nRowUsed))
+            lprintf (r, "[%d]TAB:  nResult = %d\n", r -> nPid, r -> TableStack.State.nResult) ;
 
         }
      else if (num == 0)
         {
         *pRet = NULL ;
-        if (bDebug & dbgEval)
-            lprintf ("[%d]EVAL> <NULL>\n", nPid) ;
+        if (r -> bDebug & dbgEval)
+            lprintf (r, "[%d]EVAL> <NULL>\n", r -> nPid) ;
         }
      else
         {
         *pRet = &sv_undef ;
-        if (bDebug & dbgEval)
-            lprintf ("[%d]EVAL> returns %d args instead of one\n", nPid, num) ;
+        if (r -> bDebug & dbgEval)
+            lprintf (r, "[%d]EVAL> returns %d args instead of one\n", r -> nPid, num) ;
         }
 
      /*if (SvREFCNT(*pRet) != 2)
-            lprintf ("[%d]EVAL refcnt != 2 !!= %d !!!!!\n", nPid, SvREFCNT(*pRet)) ;*/
+            lprintf (r, "[%d]EVAL refcnt != 2 !!= %d !!!!!\n", r -> nPid, SvREFCNT(*pRet)) ;*/
 
      PUTBACK;
      FREETMPS ;
      LEAVE ;
 
-
+     if (r -> bExit)
+	 {
+	 if (*pRet)
+	     SvREFCNT_dec (*pRet) ;
+	 *pRet = NULL ;
+	 return rcExit ;
+	 }
+     
      pSVErr = ERRSV ;
      if (SvTRUE (pSVErr))
         {
@@ -386,36 +405,111 @@ static int CallCV  (/*in*/  const char *  sArg,
  	     * When we get this return, we'll just give up and quit this file completely,
  	     * without error. */
              
-            struct magic * m = SvMAGIC (pSVErr) ;
+	    /*struct magic * m = SvMAGIC (pSVErr) ;*/
 
-            sv_unmagic(pSVErr,'U');
+	    sv_unmagic(pSVErr,'U');
 	    sv_setpv(pSVErr,"");
 
+	    r -> bOptions |= optNoUncloseWarn ;
+	    r -> bExit = 1 ;
+
             return rcExit ;
             }
 
         p = SvPV (pSVErr, l) ;
-        if (l > sizeof (errdat1) - 1)
-            l = sizeof (errdat1) - 1 ;
-        strncpy (errdat1, p, l) ;
-        if (l > 0 && errdat1[l-1] == '\n')
+        if (l > sizeof (r -> errdat1) - 1)
+            l = sizeof (r -> errdat1) - 1 ;
+        strncpy (r -> errdat1, p, l) ;
+        if (l > 0 && r -> errdat1[l-1] == '\n')
              l-- ;
-        errdat1[l] = '\0' ;
+        r -> errdat1[l] = '\0' ;
          
-	LogError (rcEvalErr) ;
+	LogError (r, rcEvalErr) ;
 
 	sv_setpv(pSVErr,"");
 
 	return rcEvalErr ;
         }
 
-     if (bDebug & dbgWatchScalar)
-         Watch () ;
+     if (r -> bDebug & dbgWatchScalar)
+         Watch (r) ;
 
      
     return ok ;
     }
 
+/* -------------------------------------------------------------------------------
+*
+* Eval PERL Statements and setup the correct return value/error message
+* 
+* in  sArg   Statement to eval
+* out ppSV   pointer to an SV with should be set to CV of the evaled code
+*
+------------------------------------------------------------------------------- */
+
+
+static int EvalOnly (/*i/o*/ register req * r,
+			/*in*/  const char *  sArg,
+                        /*in*/  SV **         ppSV,
+                        /*in*/  int           flags,
+  		        /*in*/  const char *  sName)
+
+
+
+    {
+    int     rc ;
+    SV *   pSub ;
+    
+    
+    EPENTRY (EvalOnly) ;
+
+    r -> lastwarn[0] = '\0' ;
+    
+    rc = EvalAll (r, sArg, flags, sName, &pSub) ;
+
+    if (rc == ok && (flags & G_DISCARD))
+	{
+	if (pSub)
+	    SvREFCNT_dec (pSub) ;
+	return ok ;
+	}
+
+    if (ppSV && *ppSV)
+	 SvREFCNT_dec (*ppSV) ;
+
+    if (rc == ok && pSub != NULL && SvTYPE (pSub) == SVt_RV)
+        {
+        /*sv_setsv (*ppSV, pSub) ;*/
+        *ppSV = SvRV(pSub) ;
+        SvREFCNT_inc (*ppSV) ;  
+        }
+    else
+        {
+        if (pSub != NULL && SvTYPE (pSub) == SVt_PV)
+            {
+	    *ppSV = pSub ; /* save error message */
+	    pSub = NULL ;
+	    }
+        else if (r -> lastwarn[0] != '\0')
+	    {
+    	    *ppSV = newSVpv (r -> lastwarn, 0) ;
+	    }
+        else
+	    {
+    	    *ppSV = newSVpv ("Compile Error", 0) ;
+	    }
+        
+        if (pSub)
+	     SvREFCNT_dec (pSub) ;
+
+	r -> bError = 1 ;
+        return rc ;
+        }
+
+    return ok ;
+    }
+
+
 /* -------------------------------------------------------------------------------
 *
 * Eval PERL Statements and execute the evaled code
@@ -427,7 +521,8 @@ static int CallCV  (/*in*/  const char *  sArg,
 ------------------------------------------------------------------------------- */
 
 
-static int EvalAndCall (/*in*/  const char *  sArg,
+static int EvalAndCall (/*i/o*/ register req * r,
+			/*in*/  const char *  sArg,
                         /*in*/  SV **         ppSV,
                         /*in*/  int           flags,
                         /*out*/ SV **         pRet)             
@@ -435,48 +530,36 @@ static int EvalAndCall (/*in*/  const char *  sArg,
 
     {
     int     rc ;
-    SV *   pSub ;
     
     
     EPENTRY (EvalAndCall) ;
 
-    lastwarn[0] = '\0' ;
-    
-    rc = EvalAll (sArg, flags, &pSub) ;
+    if ((rc = EvalOnly (r, sArg, ppSV, flags, "")) != ok)
+	{
+	*pRet = NULL ;
+	return rc ;
+	}
 
-    if (rc == ok && pSub != NULL && SvTYPE (pSub) == SVt_RV)
-        {
-        /*sv_setsv (*ppSV, pSub) ;*/
-        SvREFCNT_dec (*ppSV) ;  
-        *ppSV = SvRV(pSub) ;
-        SvREFCNT_inc (*ppSV) ;  
-        }
-    else
-        {
-        if (pSub != NULL && SvTYPE (pSub) == SVt_PV)
-            *ppSV = pSub ; /* save error message */
-        else if (lastwarn[0] != '\0')
-    	    *ppSV = newSVpv (lastwarn, 0) ;
-        else
-    	    *ppSV = newSVpv ("Compile Error", 0) ;
-        
-        *pRet = NULL ;
-        bError = 1 ;
-        return rc ;
-        }
 
     if (*ppSV && SvTYPE (*ppSV) == SVt_PVCV)
         { /* Call the compiled eval */
-        return CallCV (sArg, (CV *)*ppSV, flags, pRet) ;
+        return CallCV (r, sArg, (CV *)*ppSV, flags, pRet) ;
         }
     
     *pRet = NULL ;
-    bError = 1 ;
+    r -> bError = 1 ;
     
-    if (lastwarn[0] != '\0')
-    	*ppSV = newSVpv (lastwarn, 0) ;
+    if (ppSV && *ppSV)
+	 SvREFCNT_dec (*ppSV) ;
+
+    if (r -> lastwarn[0] != '\0')
+    	{
+ 	*ppSV = newSVpv (r -> lastwarn, 0) ;
+	}
     else
+	{
     	*ppSV = newSVpv ("Compile Error", 0) ;
+	}
 
     return rcEvalErr ;
     }
@@ -492,42 +575,42 @@ static int EvalAndCall (/*in*/  const char *  sArg,
 *
 ------------------------------------------------------------------------------- */
 
-int Eval (/*in*/  const char *  sArg,
+int Eval (/*i/o*/ register req * r,
+			/*in*/  const char *  sArg,
           /*in*/  int           nFilepos,
           /*out*/ SV **         pRet)             
 
 
     {
-    int     rc ;
     SV **   ppSV ;
     
     
     EPENTRY (Eval) ;
 
-    numEvals++ ;
+    r -> numEvals++ ;
     *pRet = NULL ;
 
-    if (bDebug & dbgCacheDisable)
-        return EvalAllNoCache (sArg, pRet) ;
+    if (r -> bDebug & dbgCacheDisable)
+        return EvalAllNoCache (r, sArg, pRet) ;
 
     /* Already compiled ? */
 
-    ppSV = hv_fetch(pCacheHash, (char *)&nFilepos, sizeof (nFilepos), 1) ;  
+    ppSV = hv_fetch(r -> Buf.pFile -> pCacheHash, (char *)&nFilepos, sizeof (nFilepos), 1) ;  
     if (ppSV == NULL)
         return rcHashError ;
 
     if (*ppSV != NULL && SvTYPE (*ppSV) == SVt_PV)
         {
-        strncpy (errdat1, SvPV(*ppSV, na), sizeof (errdat1) - 1) ; 
-        LogError (rcEvalErr) ;
+        strncpy (r -> errdat1, SvPV(*ppSV, na), sizeof (r -> errdat1) - 1) ; 
+        LogError (r, rcEvalErr) ;
         return rcEvalErr ;
         }
 
     if (*ppSV == NULL || SvTYPE (*ppSV) != SVt_PVCV)
-        return EvalAndCall (sArg, ppSV, G_SCALAR, pRet) ;
+        return EvalAndCall (r, sArg, ppSV, G_SCALAR, pRet) ;
 
-    numCacheHits++ ;
-    return CallCV (sArg, (CV *)*ppSV, G_SCALAR, pRet) ;
+    r -> numCacheHits++ ;
+    return CallCV (r, sArg, (CV *)*ppSV, G_SCALAR, pRet) ;
     }
 
 
@@ -543,63 +626,64 @@ int Eval (/*in*/  const char *  sArg,
 ------------------------------------------------------------------------------- */
 
 
-int EvalTransFlags (/*in*/  char *   sArg,
+int EvalTransFlags (/*i/o*/ register req * r,
+			/*in*/  char *   sArg,
                     /*in*/  int      nFilepos,
                     /*in*/  int      flags,
                     /*out*/ SV **    pRet)             
 
 
     {
-    int     rc ;
     SV **   ppSV ;
     
     EPENTRY (EvalTrans) ;
 
 
-    numEvals++ ;
+    r -> numEvals++ ;
     *pRet = NULL ;
 
-    if (bDebug & dbgCacheDisable)
+    if (r -> bDebug & dbgCacheDisable)
         {
         /* strip off all <HTML> Tags */
-        TransHtml (sArg) ;
+        TransHtml (r, sArg, 0) ;
         
-        return EvalAllNoCache (sArg, pRet) ;
+        return EvalAllNoCache (r, sArg, pRet) ;
         }
 
     /* Already compiled ? */
 
-    ppSV = hv_fetch(pCacheHash, (char *)&nFilepos, sizeof (nFilepos), 1) ;  
+    ppSV = hv_fetch(r -> Buf.pFile -> pCacheHash, (char *)&nFilepos, sizeof (nFilepos), 1) ;  
     if (ppSV == NULL)
         return rcHashError ;
 
     if (*ppSV != NULL && SvTYPE (*ppSV) == SVt_PV)
         {
-        strncpy (errdat1, SvPV(*ppSV, na), sizeof (errdat1) - 1) ; 
-        LogError (rcEvalErr) ;
+        strncpy (r -> errdat1, SvPV(*ppSV, na), sizeof (r -> errdat1) - 1) ; 
+        LogError (r, rcEvalErr) ;
         return rcEvalErr ;
         }
 
     if (*ppSV == NULL || SvTYPE (*ppSV) != SVt_PVCV)
         {
         /* strip off all <HTML> Tags */
-        TransHtml (sArg) ;
+        TransHtml (r, sArg, 0) ;
 
-        return EvalAndCall (sArg, ppSV, flags, pRet) ;
+        return EvalAndCall (r, sArg, ppSV, flags, pRet) ;
         }
 
-    numCacheHits++ ;
+    r -> numCacheHits++ ;
     
-    return CallCV (sArg, (CV *)*ppSV, flags, pRet) ;
+    return CallCV (r, sArg, (CV *)*ppSV, flags, pRet) ;
     }
 
 
-int EvalTrans      (/*in*/  char *   sArg,
+int EvalTrans (/*i/o*/ register req * r,
+			/*in*/  char *   sArg,
                     /*in*/  int      nFilepos,
                     /*out*/ SV **    pRet)             
     
     {
-    return EvalTransFlags (sArg, nFilepos, G_SCALAR, pRet) ;
+    return EvalTransFlags (r, sArg, nFilepos, G_SCALAR, pRet) ;
     }
     
     
@@ -616,48 +700,139 @@ int EvalTrans      (/*in*/  char *   sArg,
 ------------------------------------------------------------------------------- */
 
 
-int EvalTransOnFirstCall (/*in*/  char *   sArg,
+int EvalTransOnFirstCall (/*i/o*/ register req * r,
+			/*in*/  char *   sArg,
                           /*in*/  int      nFilepos,
                           /*out*/ SV **    pRet)             
 
 
     {
-    int     rc ;
     SV **   ppSV ;
     
     EPENTRY (EvalTrans) ;
 
 
-    numEvals++ ;
+    r -> numEvals++ ;
     *pRet = NULL ;
 
     /* Already compiled ? */
 
-    ppSV = hv_fetch(pCacheHash, (char *)&nFilepos, sizeof (nFilepos), 1) ;  
+    ppSV = hv_fetch(r -> Buf.pFile -> pCacheHash, (char *)&nFilepos, sizeof (nFilepos), 1) ;  
     if (ppSV == NULL)
         return rcHashError ;
 
     if (*ppSV != NULL && SvTYPE (*ppSV) == SVt_PV)
         {
-        strncpy (errdat1, SvPV(*ppSV, na), sizeof (errdat1) - 1) ; 
-        LogError (rcEvalErr) ;
+        strncpy (r -> errdat1, SvPV(*ppSV, na), sizeof (r -> errdat1) - 1) ; 
+        LogError (r, rcEvalErr) ;
         return rcEvalErr ;
         }
 
     if (*ppSV == NULL || SvTYPE (*ppSV) != SVt_PVCV)
         {
         /* strip off all <HTML> Tags */
-        TransHtml (sArg) ;
+        TransHtml (r, sArg, 0) ;
 
-        return EvalAndCall (sArg, ppSV, G_SCALAR, pRet) ;
+        return EvalAndCall (r, sArg, ppSV, G_SCALAR, pRet) ;
         }
 
-    numCacheHits++ ;
+    r -> numCacheHits++ ;
     
     return ok ; /* Do not call this a second time */
     }
 
 
+/* -------------------------------------------------------------------------------
+*
+* Eval PERL Statements into a sub, check if it's already compiled
+* 
+* in  sArg      Statement to eval wrap into a sub
+* in  nFilepos  position von eval in file (is used to build an unique key)
+* in  sName     sub name
+*
+------------------------------------------------------------------------------- */
+
+int EvalSub (/*i/o*/ register req * r,
+	    /*in*/  const char *  sArg,
+	    /*in*/  int           nFilepos,
+	    /*in*/  const char *  sName)
+
+
+    {
+    int     rc ;
+    SV **   ppSV ;
+    
+    
+    EPENTRY (EvalSub) ;
+
+    r -> numEvals++ ;
+
+
+    /* Already compiled ? */
+
+    ppSV = hv_fetch(r -> Buf.pFile -> pCacheHash, (char *)&nFilepos, sizeof (nFilepos), 1) ;  
+    if (ppSV == NULL)
+        return rcHashError ;
+
+    if (*ppSV != NULL && SvTYPE (*ppSV) == SVt_PV)
+        {
+        strncpy (r -> errdat1, SvPV(*ppSV, na), sizeof (r -> errdat1) - 1) ; 
+        LogError (r, rcEvalErr) ;
+        return rcEvalErr ;
+        }
+
+    if (*ppSV == NULL || SvTYPE (*ppSV) != SVt_PVCV)
+        {
+	char endc ;
+	int  len = strlen (sName) ;
+	
+	while (len > 0 && isspace(sName[len-1]))
+	    len-- ;
+	endc = sName[len] ;
+	((char *)sName)[len] = '\0' ;
+	
+	if ((rc =  EvalOnly (r, sArg, ppSV, 0, sName)) != ok)
+	    {
+	    ((char *)sName)[len] = endc ;
+	    return rc ;
+	    }
+
+        if (r -> pImportStash && *ppSV && SvTYPE (*ppSV) == SVt_PVCV)
+	    {
+	    hv_store (r -> Buf.pFile -> pExportHash, (char *)sName, len, newRV_inc(*ppSV), 0) ;
+	    
+	    if (r -> bDebug & dbgImport)
+		lprintf (r, "[%d]IMP:  %s -> %s (%x)\n", r -> nPid, sName, HvNAME (r -> pImportStash), *ppSV) ;
+
+	    /* 
+	    gvp = (GV**)hv_fetch(r -> pImportStash, (char *)sName, len, 1);
+	    
+	    if (!gvp || *gvp == (GV*)&PL_sv_undef)
+		{
+		((char *)sName)[len] = endc ;
+		return rcHashError ;
+		}
+
+	    gv = *gvp;
+	    if (SvTYPE(gv) != SVt_PVGV) 
+		gv_init(gv, r -> pImportStash, (char *)sName, len, 0);
+	    
+	    lprintf (r, "sv_any=%x\n", gv -> sv_any) ;
+	    
+	    SvREFCNT_dec (GvCV (gv)) ;  
+	    GvCV (gv) = (CV *)*ppSV ; 
+	    SvREFCNT_inc (*ppSV) ;  
+	    */
+	    }
+
+	((char *)sName)[len] = endc ;
+	return ok ;
+	}
+
+    r -> numCacheHits++ ;
+    return ok ;
+    }
+
 
 /* -------------------------------------------------------------------------------
 *
@@ -671,7 +846,8 @@ int EvalTransOnFirstCall (/*in*/  char *   sArg,
 
 
 
-int EvalNum (/*in*/  char *        sArg,
+int EvalNum (/*i/o*/ register req * r,
+			/*in*/  char *        sArg,
              /*in*/  int           nFilepos,
              /*out*/ int *         pNum)             
     {
@@ -681,7 +857,7 @@ int EvalNum (/*in*/  char *        sArg,
     EPENTRY (EvalNum) ;
 
 
-    n = Eval (sArg, nFilepos, &pRet) ;
+    n = Eval (r, sArg, nFilepos, &pRet) ;
     
     if (pRet)
         {
@@ -707,7 +883,8 @@ int EvalNum (/*in*/  char *        sArg,
 
 
 
-int EvalBool (/*in*/  char *        sArg,
+int EvalBool (/*i/o*/ register req * r,
+	      /*in*/  char *        sArg,
               /*in*/  int           nFilepos,
               /*out*/ int *         pTrue)             
     {
@@ -717,7 +894,7 @@ int EvalBool (/*in*/  char *        sArg,
     EPENTRY (EvalNum) ;
 
 
-    rc = Eval (sArg, nFilepos, &pRet) ;
+    rc = Eval (r, sArg, nFilepos, &pRet) ;
     
     if (pRet)
         {
@@ -730,3 +907,134 @@ int EvalBool (/*in*/  char *        sArg,
     return rc ;
     }
     
+
+
+/* -------------------------------------------------------------------------------
+*
+* EvalMain Scan file for [* ... *] and convert it to a perl program
+* 
+*
+------------------------------------------------------------------------------- */
+
+
+int EvalMain (/*i/o*/ register req *  r)
+
+    {
+    int     rc ;
+    long    nFilepos = -1 ;
+    char *  sProg = "" ;
+    SV **   ppSV ;
+    SV *    pRet ;
+    int     flags = G_SCALAR ;
+
+    /* Already compiled ? */
+
+    ppSV = hv_fetch(r -> Buf.pFile -> pCacheHash, (char *)&nFilepos, sizeof (nFilepos), 1) ;  
+    if (ppSV == NULL)
+        return rcHashError ;
+
+    if (*ppSV != NULL && SvTYPE (*ppSV) == SVt_PV)
+        {
+        strncpy (r -> errdat1, SvPV(*ppSV, na), sizeof (r -> errdat1) - 1) ; 
+        LogError (r, rcEvalErr) ;
+        return rcEvalErr ;
+        }
+
+    if (*ppSV == NULL || SvTYPE (*ppSV) != SVt_PVCV)
+	{ /* Not already compiled -> build a perl frame program */
+	char * pStart = r -> Buf.pBuf ;
+	char * pEnd   = r -> Buf.pEndPos ;
+	char * pOpenBracket  = r -> pConf -> pOpenBracket ;
+	char * pCloseBracket = r -> pConf -> pCloseBracket ;
+	int  lenOpenBracket  = strlen (pOpenBracket) ;
+	int  lenCloseBracket = strlen (pCloseBracket) ;
+	char * pOpen  ;
+	char * pClose ;
+	char   buf [256] ;
+        int    nBlockNo = 1 ;
+
+        
+	if (r -> sSubName && *(r -> sSubName))
+	    {
+	    int nPos = GetSubTextPos (r, r -> sSubName) ;
+	    
+	    if (!nPos || pStart + nPos > pEnd || nPos < 0)
+		{
+		strncpy (r -> errdat1, r -> sSubName, sizeof (r -> errdat1) - 1) ; 
+		return rcSubNotFound ;
+		}
+	    pStart += nPos ; 
+	    }
+	pOpen = pStart - 1 ;
+	
+	do 
+            pOpen  = strstr (pOpen + 1, pOpenBracket) ;
+        while (pOpen && pOpen > pStart && pOpen[-1] == '[') ;
+        
+        
+        if (!pOpen)
+            { /* no top level perl blocks -> call ProcessBlock directly */
+            ProcessBlock (r, pStart - r -> Buf.pBuf, r -> Buf.pEndPos - r -> Buf.pBuf, 1) ;
+            return ok ;
+            }
+
+
+	OutputToMemBuf (r, NULL, r -> Buf.pEndPos - r -> Buf.pBuf) ;
+
+	while (pStart)
+	    {
+	    pClose = NULL ;
+	    if (pOpen)
+                {
+                if ((pClose = strstr (pOpen, pCloseBracket)) == NULL)
+                    {
+                    strncpy (r -> errdat1, pCloseBracket, sizeof (r -> errdat1) - 1) ; 
+                    return rcMissingRight ;
+                    }
+                *pOpen = '\0' ;
+                }
+            else
+		pOpen = pEnd ;
+
+
+            sprintf (buf, "\n$___b=$_[0] -> ProcessBlock (%d,%d,%d);\ngoto \"b$___b\";\nb%d:;\n", pStart - r -> Buf.pBuf, pOpen - pStart, nBlockNo, nBlockNo) ;
+            oputs  (r, buf) ;
+            nBlockNo++ ;
+	    if (pClose)
+		{
+		owrite (r, pOpen + lenOpenBracket, pClose - (pOpen + lenOpenBracket)) ;
+		pStart = pClose + lenCloseBracket ;
+                /* skip trailing whitespaces */
+                while (isspace(*pStart))
+                    pStart++ ;
+                pOpen  = pStart - 1 ;
+                do 
+                    pOpen  = strstr (pOpen + 1, pOpenBracket) ;
+                while (pOpen && pOpen > pStart && pOpen[-1] == '[') ;
+                }
+	    else
+                {
+                pStart = NULL ;
+                }
+            }
+
+        oputs  (r, "\nb0:\n\0") ;
+
+	sProg = OutputToStd (r) ;
+	if (sProg == NULL)
+	    return rcOutOfMemory ;
+
+        /* strip off all <HTML> Tags */
+	TransHtml (r, sProg, 0) ;
+
+        if ((rc = EvalAndCall (r, sProg, ppSV, flags, &pRet)) != ok)
+            return rc ;
+        return ok ; /* SvIV (pRet) ;*/
+        }
+
+    r -> numCacheHits++ ;
+    
+    if ((rc = CallCV (r, sProg, (CV *)*ppSV, flags, &pRet)) != ok)
+        return rc ;
+    return ok ; /* SvIV (pRet) ;*/
+    }
