@@ -1,6 +1,6 @@
 /*###################################################################################
 #
-#   Embperl - Copyright (c) 1997-1998 Gerald Richter / ECOS
+#   Embperl - Copyright (c) 1997-1999 Gerald Richter / ECOS
 #
 #   You may distribute under the terms of either the GNU General Public
 #   License or the Artistic License, as specified in the Perl README file.
@@ -38,24 +38,6 @@
 #define lwrite                 EMBPERL_lwrite            
 #define _free                  EMBPERL__free             
 #define _malloc                EMBPERL__malloc           
-#define pReq                   EMBPERL_pReq              
-#define bDebug                 EMBPERL_bDebug            
-#define nIOType                EMBPERL_nIOType           
-#define errdat1                EMBPERL_errdat1           
-#define errdat2                EMBPERL_errdat2          
-#define Stack                  EMBPERL_Stack             
-#define nStack                 EMBPERL_nStack            
-#define State                  EMBPERL_State             
-#define ArgStack               EMBPERL_ArgStack          
-#define pArgStack              EMBPERL_pArgStack         
-#define HtmlStack              EMBPERL_HtmlStack             
-#define nHtmlStack             EMBPERL_nHtmlStack            
-#define HtmlState              EMBPERL_HtmlState             
-#define ArgHtmlStack           EMBPERL_ArgHtmlStack          
-#define pArgHtmlStack          EMBPERL_pArgHtmlStack         
-#define nTabMode               EMBPERL_nTabMode          
-#define nTabMaxRow             EMBPERL_nTabMaxRow        
-#define nTabMaxCol             EMBPERL_nTabMaxCol        
 #define LogError               EMBPERL_LogError          
 #define OutputToHtml           EMBPERL_OutputToHtml      
 #define Eval                   EMBPERL_Eval              
@@ -64,99 +46,70 @@
 #define stristr                EMBPERL_stristr           
 #define strlower               EMBPERL_strlower          
 #define TransHtml              EMBPERL_TransHtml         
+#define TransHtmlSV            EMBPERL_TransHtmlSV
 #define GetHtmlArg             EMBPERL_GetHtmlArg        
 #define GetHashValueLen        EMBPERL_GetHashValueLen   
 #define GetHashValue           EMBPERL_GetHashValue      
 #define Char2Html              EMBPERL_Char2Html         
 #define Html2Char              EMBPERL_Html2Char         
 #define sizeHtml2Char          EMBPERL_sizeHtml2Char     
-#define nPid                   EMBPERL_nPid 
 #define OutputToMemBuf         EMBPERL_OutputToMemBuf
 #define OutputToStd            EMBPERL_OutputToStd
 #define GetLogHandle           EMBPERL_GetLogHandle
-#define TableStack 	       EMBPERL_TableStack
-#define nTableStack 	       EMBPERL_nTableStack
-#define TableState 	       EMBPERL_TableState
-#define pEvalErr               EMBPERL_pEvalErr   
-#define pCacheHash             EMBPERL_pCacheHash 
-#define pNameSpace             EMBPERL_pNameSpace
-#define pBuf                   EMBPERL_pBuf          
-#define pCurrPos               EMBPERL_pCurrPos      
-#define pCurrStart             EMBPERL_pCurrStart    
-#define pEndPos                EMBPERL_pEndPos       
-#define pCurrTag               EMBPERL_pCurrTag      
-#define pEnvHash               EMBPERL_pEnvHash      
-#define pFormHash              EMBPERL_pFormHash     
-#define pFormSplitHash         EMBPERL_pFormSplitHash     
-#define pInputHash             EMBPERL_pInputHash    
-#define pNameSpaceHash         EMBPERL_pNameSpaceHash
-#define pFormArray             EMBPERL_pFormArray    
-#define cMultFieldSep          EMBPERL_cMultFieldSep 
 #define SearchCmd              EMBPERL_SearchCmd     
 #define ProcessCmd             EMBPERL_ProcessCmd    
-#define bSafeEval              EMBPERL_bSafeEval
-#define sLogfileURLName        EMBPERL_sLogfileURLName
-#define numEvals               EMBPERL_numEvals
-#define numCacheHits           EMBPERL_numCacheHits 
+#define ProcessSub             EMBPERL_ProcessSub
 #define Char2Url               EMBPERL_Char2Url            
 #define CmdTab                 EMBPERL_CmdTab              
 #define EvalTrans              EMBPERL_EvalTrans           
+#define EvalMain               EMBPERL_EvalMain
 #define EvalTransFlags         EMBPERL_EvalTransFlags
 #define EvalTransOnFirstCall   EMBPERL_EvalTransOnFirstCall           
+#define EvalSub                EMBPERL_EvalSub
 #define GetContentLength       EMBPERL_GetContentLength    
 #define GetLogFilePos          EMBPERL_GetLogFilePos       
 #define ReadHTML               EMBPERL_ReadHTML            
 #define ScanCmdEvalsInString   EMBPERL_ScanCmdEvalsInString
-#define bEscMode               EMBPERL_bEscMode            
-#define bEscModeSet            EMBPERL_bEscModeSet            
-#define nAllocSize             EMBPERL_nAllocSize          
-#define pCurrEscape            EMBPERL_pCurrEscape         
-#define pNextEscape            EMBPERL_pNextEscape         
-#define bStrict                EMBPERL_bStrict
 #define EvalDirect             EMBPERL_EvalDirect
-#define bOptions               EMBPERL_bOptions    
-#define bReqRunning            EMBPERL_bReqRunning 
-#define nEvalPackage           EMBPERL_nEvalPackage
-#define pEvalPackage           EMBPERL_pEvalPackage
-#define sEvalPackage           EMBPERL_sEvalPackage
-#define pPackage               EMBPERL_pPackage    
-#define pErrArray              EMBPERL_pErrArray   
-#define pOpcodeMask            EMBPERL_pOpcodeMask 
-#define sSourcefile            EMBPERL_sSourcefile
-#define nSourceline            EMBPERL_nSourceline
-#define pSourcelinePos         EMBPERL_nSourcelinePos
 #define GetLineNo              EMBPERL_GetLineNo
-#define pLineNoCurrPos         EMBPERL_pLineNoCurrPos
-#define bError                 EMBPERL_bError
-#define lastwarn               EMBPERL_lastwarn
-#define lastwarn               EMBPERL_lastwarn
 #define Dirname                EMBPERL_Dirname
-#define nMarker                EMBPERL_nMarker
 #define CommitError            EMBPERL_CommitError
 #define RollbackError          EMBPERL_RollbackError
-#define pErrFill	       EMBPERL_pErrFill
-#define pErrState              EMBPERL_pErrState
 #define _memstrcat             EMBPERL__memstrcat
+#define _ep_strdup             EMBPERL__ep_strdup
+#define _ep_strndup            EMBPERL__ep_strndup
 #define _realloc               EMBPERL__realloc
+#define ExecuteReq             EMBPERL_ExecuteReq     
+#define FreeConfData           EMBPERL_FreeConfData   
+#define FreeRequest            EMBPERL_FreeRequest    
+#define GetHashValueInt        EMBPERL_GetHashValueInt
+#define GetHashValueStr        EMBPERL_GetHashValueStr
+#define Init                   EMBPERL_Init           
+#define ResetHandler           EMBPERL_ResetHandler   
+#define SetupConfData          EMBPERL_SetupConfData  
+#define SetupFileData          EMBPERL_SetupFileData  
+#define SetupRequest           EMBPERL_SetupRequest   
+#define Term                   EMBPERL_Term           
+#define sstrdup                EMBPERL_sstrdup        
+#define ProcessBlock           EMBPERL_ProcessBlock
+#define NewEscMode             EMBPERL_NewEscMode
+#define GetSubTextPos          EMBPERL_GetSubTextPos
+#define SetSubTextPos          EMBPERL_SetSubTextPos
+#define SetupDebugger          EMBPERL_SetupDebugger
 
 
-/* --> from mod_perl.h
- * patchlevel.h causes a -Wall warning, 
- * plus chance that another patchlevel.h might be in -I paths
- * so try to avoid it if possible 
- */ 
-#ifdef PERL_VERSION
-#if PERL_VERSION >= 500476
-#define PERL5_005
+#define InitialReq             EMBPERL_InitialReq
+#define pCurrReq               EMBPERL_pCurrReq
+
+#ifndef PERL_VERSION
+#include <patchlevel.h>
+#define PERL_VERSION PATCHLEVEL
+#define PERL_SUBVERSION SUBVERSION
 #endif
-#else
-#include "patchlevel.h"
-#if ((PATCHLEVEL >= 4) && (SUBVERSION >= 76)) || (PATCHLEVEL >= 5)
-#define PERL5_005
-#endif
-#endif /*PERL_VERSION*/
 
-#ifdef PERL5_005
+
+#if PERL_VERSION >= 5
+
 #ifndef rs
 #define rs PL_rs
 #endif
@@ -223,10 +176,31 @@
 #ifndef maxo
 #define maxo PL_maxo
 #endif
-#else
 
+#if PERL_SUBVERSION >= 50 || PERL_VERSION >= 6
+
+#ifndef na
+#define na PL_na
+#endif
+#ifndef sv_undef
+#define sv_undef PL_sv_undef
+#endif
+#ifndef tainted
+#define tainted PL_tainted
+#endif
+
+#endif
+
+
+#else  /* PERL_VERSION > 5 */
+
+#ifndef ERRSV
 #define ERRSV GvSV(errgv)
+#endif
+
+#ifndef dTHR
 #define dTHR
+#endif
 
 
-#endif /* PERL5_005 */
+#endif /* PERL_VERSION > 5 */
