@@ -1,6 +1,6 @@
 /*###################################################################################
 #
-#   Embperl - Copyright (c) 1997-1998 Gerald Richter / ECOS
+#   Embperl - Copyright (c) 1997-1999 Gerald Richter / ECOS
 #
 #   You may distribute under the terms of either the GNU General Public
 #   License or the Artistic License, as specified in the Perl README file.
@@ -19,13 +19,15 @@
 
 
 /* ---------------------------------------------------------------------------- */
-/* Output a string and escapte html special character to html special representation (&xxx;) */
-/* */
-/* i/o sData     = input:  perl string */
-/* */
+/* Output a string and escape html special character to html special            */
+/* representation (&xxx;)                                                       */
+/*                                                                              */
+/* i/o sData     = input:  perl string                                          */
+/*                                                                              */
 /* ---------------------------------------------------------------------------- */
 
-void OutputToHtml (/*i/o*/ const char *  sData)
+void OutputToHtml (/*i/o*/ register req * r,
+ 		   /*i/o*/ const char *   sData)
 
     {
     char * pHtml  ;
@@ -33,9 +35,9 @@ void OutputToHtml (/*i/o*/ const char *  sData)
     
     EPENTRY (OutputToHtml) ;
 
-    if (pCurrEscape == NULL)
+    if (r -> pCurrEscape == NULL)
         {
-        oputs (sData) ;
+        oputs (r, sData) ;
         return ;
         }
 
@@ -45,38 +47,38 @@ void OutputToHtml (/*i/o*/ const char *  sData)
         if (*sData == '\\')
             {
             if (p != sData)
-                owrite (p, 1, sData - p) ;
+                owrite (r, p, sData - p) ;
             sData++ ;
             p = sData ;
             }
         else
             {
-            pHtml = pCurrEscape[(unsigned char)(*sData)].sHtml ;
+            pHtml = r -> pCurrEscape[(unsigned char)(*sData)].sHtml ;
             if (*pHtml)
                 {
                 if (p != sData)
-                    owrite (p, 1, sData - p) ;
-                oputs (pHtml) ;
+                    owrite (r, p, sData - p) ;
+                oputs (r, pHtml) ;
                 p = sData + 1;
                 }
             }
         sData++ ;
         }
     if (p != sData)
-        owrite (p, 1, sData - p) ;
+        owrite (r, p, sData - p) ;
     }
 
 
-
+#if 0
 
 /* ---------------------------------------------------------------------------- */
-/* find substring ignore case */
-/* */
-/* in  pSring  = string to search in (any case) */
-/* in  pSubStr = string to search for (must be upper case) */
-/* */
-/* out ret  = pointer to pSubStr in pStringvalue or NULL if not found */
-/* */
+/* find substring ignore case                                                   */
+/*                                                                              */
+/* in  pSring  = string to search in (any case)                                 */
+/* in  pSubStr = string to search for (must be upper case)                      */
+/*                                                                              */
+/* out ret  = pointer to pSubStr in pStringvalue or NULL if not found           */
+/*                                                                              */
 /* ---------------------------------------------------------------------------- */
 
 
@@ -126,6 +128,31 @@ static char * strlower (/*in*/ char *   pString)
     return pString ;
     }
 
+#endif
+
+/* ---------------------------------------------------------------------------- */
+/* save strdup                                                                  */
+/*                                                                              */
+/* in  pSring  = string to save on memory heap                                  */
+/*                                                                              */
+/* ---------------------------------------------------------------------------- */
+
+
+char * sstrdup (/*in*/ char *   pString)
+
+    {
+    char * p ;
+    
+    if (pString == NULL)
+        return NULL ;
+
+    p = malloc (strlen (pString) + 1) ;
+
+    strcpy (p, pString) ;
+
+    return p ;
+    }
+
 
 
 /* */
@@ -153,20 +180,24 @@ static int CmpCharTrans (/*in*/ const void *  pKey,
 /* 								             */
 /* i/o sData     = input:  html string                                       */
 /*                 output: perl string                                       */
+/* in  nLen      = length of sData on input (or 0 for null terminated)       */
 /*                                                                           */
 /* ------------------------------------------------------------------------- */
 
-void TransHtml (/*i/o*/ char *  sData)
+int   TransHtml (/*i/o*/ register req * r,
+		/*i/o*/ char *         sData,
+		/*in*/   int           nLen)
 
     {
     char * p = sData ;
     char * s ;
     char * e ;
     struct tCharTrans * pChar ;
+    int  bInUrl = r -> bEscInUrl ;
 
     EPENTRY (TransHtml) ;
 	
-    if (bOptions & optRawInput)
+    if (r -> bOptions & optRawInput)
 	{ /* Just remove CR for raw input */
 	while (*p != '\0')
 	    {
@@ -175,95 +206,133 @@ void TransHtml (/*i/o*/ char *  sData)
 	    p++ ;
 	    }	
 
-    	return ; 	
+	return nLen ; 	
         }
         
     s = NULL ;
-    e = sData + strlen (sData) ;
+    if (nLen == 0)
+        nLen = strlen (sData) ;
+    e = sData + nLen ;
 
-    while (*p)
-        {
-        if (*p == '\\')
-            {
-            if (p[1] == '<')
-                { /*  Quote next HTML tag */
-                p += 2 ;
-                while (*p && *p != '>')
-                    p++ ;
-                }
-            else if (p[1] == '&')
-                { /*  Quote next HTML char */
-                p += 2 ;
-                while (*p && *p != ';')
-                    p++ ;
-                }
-            else
-                p++ ; /* Nothing to quote */
-            }
-        else if (*p == '\r')
-            {
-            *p++ = ' ' ;
-            }
-        else
-            {
-            if (p[0] == '<' && (isalpha (p[1]) || p[1] == '/'))
-                { /*  count HTML tag length */
-                s = p ;
-                p++ ;
-                while (*p && *p != '>')
-                    p++ ;
-                if (*p)
-                    p++ ;
-                else
-                    { /* missing left '>' -> no html tag  */
-                    p = s ;
-                    s = NULL ;
-                    }
-                }
-            else if (p[0] == '&')
-                { /*  count HTML char length */
-                s = p ;
-                p++ ;
-                while (*p && *p != ';')
-                    p++ ;
+    while (*p && p < e)
+	{
+	if (*p == '\\')
+	    {
+        
+	    if (p[1] == '<')
+		{ /*  Quote next HTML tag */
+		memmove (p, p + 1, e - p - 1) ;
+		e[-1] = ' ' ;
+		p++ ;
+		while (*p && *p != '>')
+		    p++ ;
+		}
+	    else if (p[1] == '&')
+		{ /*  Quote next HTML char */
+		memmove (p, p + 1, e - p - 1) ;
+		e[-1] = ' ' ;
+		p++ ;
+		while (*p && *p != ';')
+		    p++ ;
+		}
+	    else if (bInUrl && p[1] == '%')
+		{ /*  Quote next URL escape */
+		memmove (p, p + 1, e - p - 1) ;
+		e[-1] = ' ' ;
+		p += 3 ;
+		}
+	    else
+		p++ ; /* Nothing to quote */
+	    }
+	else if (*p == '\r')
+	    {
+	    *p++ = ' ' ;
+	    }
+	else
+	    {
+	    if (p[0] == '<' && (isalpha (p[1]) || p[1] == '/'))
+		{ /*  count HTML tag length */
+		s = p ;
+		p++ ;
+		while (*p && *p != '>')
+		    p++ ;
+		if (*p)
+		    p++ ;
+		else
+		    { /* missing left '>' -> no html tag  */
+		    p = s ;
+		    s = NULL ;
+		    }
+		}
+	    else if (p[0] == '&')
+		{ /*  count HTML char length */
+		s = p ;
+		p++ ;
+		while (*p && *p != ';')
+		    p++ ;
 
-                if (*p)
-                    {
-                    *p = '\0' ;
-                    p++ ;
-                    pChar = (struct tCharTrans *)bsearch (s, Html2Char, sizeHtml2Char, sizeof (struct tCharTrans), CmpCharTrans) ;
-                    if (pChar)
-                        *s++ = pChar -> c ;
-                    else
-                        {
+		if (*p)
+		    {
+		    *p = '\0' ;
+		    p++ ;
+		    pChar = (struct tCharTrans *)bsearch (s, Html2Char, sizeHtml2Char, sizeof (struct tCharTrans), CmpCharTrans) ;
+		    if (pChar)
+			*s++ = pChar -> c ;
+		    else
+			{
 			*(p-1)=';' ;
-                        p = s ;
-                        s = NULL ;
-                        }
-                    }
-                else
-                    {
-                    p = s ;
-                    s = NULL ;
-                    }
-                }
+			p = s ;
+			s = NULL ;
+			}
+		    }
+		else
+		    {
+		    p = s ;
+		    s = NULL ;
+		    }
+		}
+	    else if (bInUrl && p[0] == '%' && isdigit (p[1]) && isxdigit (p[2]))
+		{ 
 
-            if (s && (p - s) > 0)
-                { /* copy rest of string, pad with spaces */
-                memmove (s, p, e - p + 1) ;
-                memset (e - (p - s), ' ', (p - s)) ;
-                p = s ;
-                s = NULL ;
-                }
-            else
-                if (*p)
-                    p++ ;
-            
-            }
-        }
+		s = p ;
+		p += 3 ;
+                *s++ = ((toupper (p[-2]) - (isdigit (p[-2])?'0':('A' - 10))) << 4) 
+                      + (toupper (p[-1]) - (isdigit (p[-1])?'0':('A' - 10)))  ;
+		}
+	    if (s && (p - s) > 0)
+		{ /* copy rest of string, pad with spaces */
+		memmove (s, p, e - p + 1) ;
+		memset (e - (p - s), ' ', (p - s)) ;
+		nLen -= p - s ;
+		p = s ;
+		s = NULL ;
+		}
+	    else
+		if (*p)
+		    p++ ;
+	    }
+	}
+
+    return nLen ;
     }
 
-                        
+
+
+void TransHtmlSV (/*i/o*/ register req * r,
+		  /*i/o*/ SV *           pSV) 
+
+    {
+    STRLEN vlen ;
+    STRLEN nlen ;
+    char * pVal = SvPV (pSV, vlen) ;
+
+    nlen = TransHtml (r, pVal, vlen) ;
+
+    pVal[nlen] = '\0' ;
+    SvCUR_set(pSV, nlen) ;
+    }		  
+
+
 /* ---------------------------------------------------------------------------- */
 /* get argument from html tag  */
 /* */
@@ -282,10 +351,9 @@ const char * GetHtmlArg (/*in*/  const char *    pTag,
     {
     const char * pVal ;
     const char * pEnd ;
-    const char * pName ;
     int l ;
 
-    EPENTRY (GetHtmlArg) ;
+    /*EPENTRY (GetHtmlArg) ;*/
 
     l = strlen (pArg) ;
     while (*pTag)
@@ -311,15 +379,36 @@ const char * GetHtmlArg (/*in*/  const char *    pTag,
             pEnd = pVal ;
             if (*pVal == '"' || *pVal == '\'')
                 {
+		char nType = '\0';
                 char q = *pVal++ ;
                 pEnd++ ;
-                while (*pEnd && *pEnd != q)
+
+		while ((*pEnd != q || nType) && *pEnd != '\0')
+		    {
+		    if (nType == '\0' && *pEnd == '[' && (pEnd[1] == '+' || pEnd[1] == '-' || pEnd[1] == '$' || pEnd[1] == '!' || pEnd[1] == '#'))
+			nType = *++pEnd ;
+		    else if (nType && *pEnd == nType && pEnd[1] == ']')
+			{
+			nType = '\0';
+			pEnd++ ;
+			}
                     pEnd++ ;
-                }
+		    }
+		}
             else
                 {
-                while (*pEnd && !isspace (*pEnd) && *pEnd != '>')
+		char nType = '\0';
+		while ((!isspace (*pEnd) || nType) && *pEnd != '\0'  && *pEnd != '>')
+		    {
+		    if (nType == '\0' && *pEnd == '[' && (pEnd[1] == '+' || pEnd[1] == '-' || pEnd[1] == '$' || pEnd[1] == '!' || pEnd[1] == '#'))
+			nType = *++pEnd ;
+		    else if (nType && *pEnd == nType && pEnd[1] == ']')
+			{
+			nType = '\0';
+			pEnd++ ;
+			}
                     pEnd++ ;
+		    }
                 }
 
             *pLen = pEnd - pVal ;
@@ -362,7 +451,7 @@ char * GetHashValueLen (/*in*/  HV *           pHash,
     char *  p ;
     STRLEN  len ;        
 
-    EPENTRY (GetHashValueLen) ;
+    /*EPENTRY (GetHashValueLen) ;*/
 
     ppSV = hv_fetch(pHash, (char *)sKey, nLen, 0) ;  
     if (ppSV != NULL)
@@ -390,6 +479,42 @@ char * GetHashValue (/*in*/  HV *           pHash,
     }
 
 
+
+
+int    GetHashValueInt (/*in*/  HV *           pHash,
+                        /*in*/  const char *   sKey,
+                        /*in*/  int            nDefault)
+
+    {
+    SV **   ppSV ;
+
+    /*EPENTRY (GetHashValueInt) ;*/
+
+    ppSV = hv_fetch(pHash, (char *)sKey, strlen (sKey), 0) ;  
+    if (ppSV != NULL)
+        return SvIV (*ppSV) ;
+        
+    return nDefault ;
+    }
+
+
+char * GetHashValueStr (/*in*/  HV *           pHash,
+                        /*in*/  const char *   sKey,
+                        /*in*/  char *         sDefault)
+
+    {
+    SV **   ppSV ;
+
+    /*EPENTRY (GetHashValueInt) ;*/
+
+    ppSV = hv_fetch(pHash, (char *)sKey, strlen (sKey), 0) ;  
+    if (ppSV != NULL)
+        return SvPV (*ppSV, na) ;
+        
+    return sDefault ;
+    }
+
+
 /* ------------------------------------------------------------------------- */
 /*                                                                           */
 /* GetLineNo								     */
@@ -402,44 +527,47 @@ char * GetHashValue (/*in*/  HV *           pHash,
 /* ------------------------------------------------------------------------- */
 
 
-int GetLineNo ()
+int GetLineNo (/*i/o*/ register req * r)
 
     {
-    char * pPos = pCurrPos ;
+    char * pPos = r -> Buf.pCurrPos ;
     
-    if (pSourcelinePos == NULL)
-        return nSourceline = 1 ;
+    if (r -> Buf.pSourcelinePos == NULL)
+	if (r -> Buf.pFile == NULL)
+	    return 0 ;
+	else
+	    return r -> Buf.nSourceline = r -> Buf.pFile -> nFirstLine ;
 
-    if (pLineNoCurrPos)
-        pPos = pLineNoCurrPos ;
+    if (r -> Buf.pLineNoCurrPos)
+        pPos = r -> Buf.pLineNoCurrPos ;
 
-    if (pPos == NULL || pPos == pSourcelinePos || pPos < pBuf || pPos > pEndPos)
-        return nSourceline ;
+    if (pPos == NULL || pPos == r -> Buf.pSourcelinePos || pPos < r -> Buf.pBuf || pPos > r -> Buf.pEndPos)
+        return r -> Buf.nSourceline ;
 
 
-    if (pPos > pSourcelinePos)
+    if (pPos > r -> Buf.pSourcelinePos)
         {
-        char * p = pSourcelinePos ;
+        char * p = r -> Buf.pSourcelinePos ;
 
-        while (p < pPos && p < pEndPos)
+        while (p < pPos && p < r -> Buf.pEndPos)
             {
             if (*p++ == '\n')
-                nSourceline++ ;
+                r -> Buf.nSourceline++ ;
             }
         }
     else
         {
-        char * p = pSourcelinePos ;
+        char * p = r -> Buf.pSourcelinePos ;
 
-        while (p > pPos && p > pBuf)
+        while (p > pPos && p > r -> Buf.pBuf)
             {
             if (*--p == '\n')
-                nSourceline-- ;
+                r -> Buf.nSourceline-- ;
             }
         }
 
-    pSourcelinePos = pPos ;
-    return nSourceline ;
+    r -> Buf.pSourcelinePos = pPos ;
+    return r -> Buf.nSourceline ;
     }
 
 
@@ -475,3 +603,107 @@ void Dirname (/*in*/ const char * filename,
 
     return ;
     }
+
+
+
+/* ------------------------------------------------------------------------- */
+/*                                                                           */
+/* GetSubTextPos						             */
+/*                                                                           */
+/*                                                                           */
+/* in	sName = name of sub                                                  */
+/*                                                                           */
+/* returns the position within the file for a given Embperl sub              */
+/*                                                                           */
+/* ------------------------------------------------------------------------- */
+
+
+int GetSubTextPos (/*i/o*/ register req * r,
+		   /*in*/  const char *   sName) 
+
+    {
+    SV **   ppSV ;
+    const char *  sKey  ;
+    char    sKeyBuf[sizeof (int) + 4] ;
+    int	    l ;    
+    
+    EPENTRY (Eval) ;
+
+    while (isspace(*sName))
+	sName++ ;
+
+    l = strlen (sName) ;
+    while (l > 0 && isspace(sName[l-1]))
+	l-- ;
+    
+    sKey = sName ;
+    if (l < sizeof (int))
+	{ /* right pad name with spaces to make sure name is longer then sizeof (int) */
+	  /* distiguish it from filepos entrys */
+	memset (sKeyBuf, ' ', sizeof (sKeyBuf) - 1) ;
+	sKeyBuf[sizeof(sKeyBuf) - 1] = '\0' ;
+	memcpy (sKeyBuf, sName, l) ;
+	sKey = sKeyBuf ;
+	l = sizeof(sKeyBuf) - 1 ;
+	}
+    
+    
+    ppSV = hv_fetch(r -> Buf.pFile -> pCacheHash, (char *)sKey, l, 0) ;  
+    if (ppSV == NULL || *ppSV == NULL) /* || SvTYPE (*ppSV) != SVt_IV)*/
+        return 0 ;
+
+    return SvIV (*ppSV) ;    
+    }
+
+
+/* ------------------------------------------------------------------------- */
+/*                                                                           */
+/* SetSubTextPos						             */
+/*                                                                           */
+/*                                                                            */
+/* in	sName = name of sub                                                  */
+/* in   nPos  = position within the file for a given Embperl sub             */
+/*                                                                           */
+/* ------------------------------------------------------------------------- */
+
+
+int SetSubTextPos (/*i/o*/ register req * r,
+		   /*in*/  const char *   sName,
+		   /*in*/  int		  nPos) 
+
+    {
+    SV **   ppSV ;
+    const char *  sKey ;
+    char    sKeyBuf[sizeof (int) + 4] ;
+    int	    l ;    
+    
+    EPENTRY (Eval) ;
+
+    while (isspace(*sName))
+	sName++ ;
+
+    l = strlen (sName) ;
+    while (l > 0 && isspace(sName[l-1]))
+	l-- ;
+    
+    sKey = sName ;
+    if (l < sizeof (int))
+	{ /* right pad name with spaces to make sure name is longer then sizeof (int) */
+	  /* distiguish it from filepos entrys */
+	memset (sKeyBuf, ' ', sizeof (sKeyBuf) - 1) ;
+	sKeyBuf[sizeof(sKeyBuf) - 1] = '\0' ;
+	memcpy (sKeyBuf, sName, l) ;
+	sKey = sKeyBuf ;
+	l = sizeof(sKeyBuf) - 1 ;
+	}
+    
+    
+    ppSV = hv_fetch(r -> Buf.pFile -> pCacheHash, (char *)sKey, l, 1) ;  
+    if (ppSV == NULL)
+        return rcHashError ;
+
+    SvREFCNT_dec (*ppSV) ;  
+    *ppSV = newSViv (nPos) ;
+    
+    return ok ;
+    }
\ No newline at end of file
