@@ -2,23 +2,6 @@
 # Before `make install' is performed this script should be runnable with
 # `make test'. After `make install' it should work as `perl test.pl'
 
-# avoid some warnings:
-
-use vars qw ($httpconfsrc $httpconf $EPPORT $EPPORT2 *SAVEERR *ERR $EPHTTPDDLL $EPSTARTUP $EPDEBUG) ;
-
-    {
-    local $^W = 0 ;
-    eval " use Win32::Process; " ;
-    $win32loaderr = $@ ;
-    eval " use Win32; " ;
-    $win32loaderr ||= $@ ;
-    }
-
-BEGIN { $fatal = 1 ; $^W = 1 ; }
-END     { 
-        print "\nTest terminated with fatal error\n" if ($fatal) ; 
-	system "kill `cat $tmppath/httpd.pid` 2> /dev/null" if ($EPHTTPD ne '' && $killhttpd && !$EPWIN32) ;
-        }
 
 @tests = (
     'ascii',
@@ -26,42 +9,67 @@ END     {
     'plain.htm',
     'plain.htm',
     'plain.htm',
-    'error.htm???7',
-    'error.htm???7',
-    'error.htm???7',
-    'noerr/noerrpage.htm???7?2',
+    'plainblock.htm',
+    'plainblock.htm',
+    'error.htm???8',
+    'error.htm???8',
+    'error.htm???8',
+    'unclosed.htm???1',
+#    'errorright.htm???1',
+    'notfound.htm???1',
+    'notallow.xhtm???1',
+    'noerr/noerrpage.htm???8?2',
+    'errdoc/errdoc.htm???8?262144',
     'rawinput/rawinput.htm????16',
     'var.htm',
     'varerr.htm???-1',
     'varerr.htm???2',
     'escape.htm',
+    'spaces.htm',
     'tagscan.htm',
     'tagscan.htm??1',
     'if.htm',
-    'loop.htm?erstes=Hallo&zweites=Leer+zeichen&drittes=%21%22%23&erstes=Wert2',
+    'ifperl.htm',
+    'loop.htm?erstes=Hallo&zweites=Leer+zeichen&drittes=%21%22%23%2a%2B&erstes=Wert2',
+    'loopperl.htm?erstes=Hallo&zweites=Leer+zeichen&drittes=%21%22%23&erstes=Wert2',
     'table.htm',
     'table.htm??1',
-    'table.htm??32769',
-#   'table.htm??131085',
     'lists.htm?sel=2&SEL1=B&SEL3=D&SEL4=cc',
+    'mix.htm',
     'nesting.htm',
     'object.htm???2',
     'discard.htm???12',
-    'input.htm?feld5=Wert5&feld6=Wert6&feld7=Wert7&feld8=Wert8&cb5=cbv5&cb6=cbv6&cb7=cbv7&cb8=cbv8&cb9=ncbv9&cb10=ncbv10&cb11=ncbv11&mult=Wert3&mult=Wert6',
+    'input.htm?feld5=Wert5&feld6=Wert6&feld7=Wert7&feld8=Wert8&cb5=cbv5&cb6=cbv6&cb7=cbv7&cb8=cbv8&cb9=ncbv9&cb10=ncbv10&cb11=ncbv11&mult=Wert3&mult=Wert6&esc=a<b&escmult=a>b&escmult=Wert3',
     'hidden.htm?feld1=Wert1&feld2=Wert2&feld3=Wert3&feld4=Wert4',
     'java.htm',
     'inputjava.htm',
     'post.htm',
-    'upload.htm',
+    'upload.htm?multval=A&multval=B&multval=C&single=S',
     'reqrec.htm',
     'reqrec.htm',
+    'include.htm',
+    'includeerr1.htm???1',
+    'includeerr2.htm???4',
     'registry/Execute.htm',
-    'registry/errpage.htm???14',
+    'registry/errpage.htm???16',
+    'registry/tied.htm???3',
+    'registry/tied.htm???3',
+    'callsub.htm',
+    'callsub.htm',
+    'importsub.htm',
+    'importsub.htm',
+    'importsub2.htm',
+    'importmodule.htm',
+    'recursexec.htm',
     'nph/div.htm????64',
-    'nph/npherr.htm???7?64',
+    'nph/npherr.htm???8?64',
+    'nph/nphinc.htm????64',
+    'sub.htm',
     'sub.htm',
     'exit.htm',
     'exit2.htm',
+    'exit3.htm',
+    'chdir.htm?a=1&b=2&c=&d=&f=5&g&h=7&=8&=',
     'chdir.htm?a=1&b=2&c=&d=&f=5&g&h=7&=8&=',
     'allform/allform.htm?a=1&b=2&c=&d=&f=5&g&h=7&=8&=???8192',
     'stdout/stdout.htm????16384',
@@ -77,10 +85,69 @@ END     {
     'safe/safe.htm???-1?4',
     'opmask/opmask.htm???-1?12?TEST',
     'opmask/opmasktrap.htm???2?12?TEST',
+    'mdatsess.htm?cnt=0',
+    'setsess.htm?a=1',
+    'mdatsess.htm?cnt=1',
+    'getnosess.htm?nocookie=2',
+    'mdatsess.htm?cnt=2',
+    'getsess.htm',
+    'mdatsess.htm?cnt=3',
+    'execgetsess.htm',
+    'clearsess.htm',
     ) ;
 
 
-$confpath = './test/conf' ;
+# avoid some warnings:
+
+use vars qw ($httpconfsrc $httpconf $EPPORT $EPPORT2 *SAVEERR *ERR $EPHTTPDDLL $EPSTARTUP $EPDEBUG
+             $EPSESSIONDS $EPSESSIONCLASS $EPSESSIONVERSION) ;
+
+    {
+    local $^W = 0 ;
+    eval " use Win32::Process; " ;
+    $win32loaderr = $@ ;
+    eval " use Win32; " ;
+    $win32loaderr ||= $@ ;
+    }
+
+BEGIN 
+    { 
+    $fatal  = 1 ;
+    $^W     = 1 ;
+    $|      = 1;
+    
+    eval 'use ExtUtils::testlib' if (defined ($ARGV[0]) && $ARGV[0] =~ /b/) ;
+
+    #### install handler which kill httpd when terminating ####
+
+    $SIG{__DIE__} = sub { 
+	return unless $_[0] =~ /^\*\*\*/ ;
+	return unless $killhttpd ;
+	if ($EPWIN32)
+	    {
+	    $HttpdObj->Kill(-1) if ($HttpdObj) ;
+	    }
+	else
+	    {
+	    system "kill `cat $tmppath/httpd.pid` 2> /dev/null" if ($EPHTTPD ne '') ;
+	    }
+	} ;
+
+    print "\nloading...                    ";
+    
+    }
+
+END 
+    { 
+    print "\nTest terminated with fatal error\n" if ($fatal) ; 
+    system "kill `cat $tmppath/httpd.pid` 2> /dev/null" if ($EPHTTPD ne '' && $killhttpd && !$EPWIN32) ;
+    exit ($fatal || $err) ;	
+    }
+
+
+
+
+$confpath = 'test/conf' ;
 
 $cmdarg   = $ARGV[0] || '' ;
 shift @ARGV ;
@@ -100,14 +167,16 @@ else
 
 
 $EPPORT2 = ($EPPORT || 0) + 1 ;
+$EPSESSIONCLASS = $ENV{EMBPERL_SESSION_CLASS} || (($EPSESSIONVERSION =~ /^0\.17/)?'Win32':'0')  || ($EPSESSIONVERSION > 1.00?'Embperl':'0') ;
+$EPSESSIONDS    = $ENV{EMBPERL_SESSION_DS} || 'dbi:mysql:session' ;
 
 die "You must install libwin32 first" if ($EPWIN32 && $win32loaderr && $EPHTTPD) ;
 
 #### setup paths #####
 
-$inpath  = './test/html' ;
-$tmppath = './test/tmp' ;
-$cmppath = './test/cmp' ;
+$inpath  = 'test/html' ;
+$tmppath = 'test/tmp' ;
+$cmppath = 'test/cmp' ;
 
 
 #### setup files ####
@@ -124,7 +193,7 @@ $logfile    = "$tmppath/test.log" ;
 $embploc = 'embperl/' ;
 if ($EPWIN32)
     {
-    $cgiloc  = 'cgi-bin-32/' ;
+    $cgiloc  = 'cgi-bin/' ; #'cgi-bin-32/' ;
     }
 else
     {
@@ -135,12 +204,15 @@ else
 $port    = $EPPORT ;
 $host    = 'localhost' ;
 $httpdpid = 0 ;
-$defaultdebug = 0x85ffd ;
+$defaultdebug = 0x785ffd ;
 
 
 if ($cmdarg =~ /\?/)
     {
     print "\n\n" ;
+    print "test.pl [options] [files]\n" ;
+    print "files: <filename>|<testnumber>|-<testnumber>\n\n" ;
+    print "options:\n" ;
     print "o	test offline\n" ;
     print "c	test cgi\n" ;
     print "h	test mod_perl\n" ;
@@ -149,13 +221,15 @@ if ($cmdarg =~ /\?/)
     print "l	loop forever\n" ;
     print "m	start httpd with mulitple childs\n" ;
     print "v    memory check\n" ;
-    print "g    exit if httpd grows after 2 loop\n" ;	
+    print "g    exit if httpd grows after 2 loop\n" ;   
     print "f    file to use for config.pl\n" ;
     print "x    do not start httpd\n" ;
     print "u    use unique filenames\n" ;
     print "n    do not check httpd errorlog\n" ;
     print "q    set debug to 0\n" ;
     print "i    ignore errors\n" ;
+    print "t    list tests\n" ;
+    print "b    use uninstalled version (from blib/..)\n" ;
     print "\n\n" ;
     print "path\t$EPPATH\n" ;
     print "httpd\t$EPHTTPD\n" ;
@@ -163,7 +237,19 @@ if ($cmdarg =~ /\?/)
     exit (1) ;
     }
 
+if ($cmdarg =~ /t/)
+    {
+    $i = 0 ;
+    foreach $t (@tests)
+	{
+	print "$i = $t\n" ;
+	$i++ ;
+	}
+    exit (1) ;
+    }
 
+
+	
 $killhttpd = 1 ; # kill httpd at end of test
 $multhttpd = 0 ; # start httpd with child fork
 $looptest  = 0 ; # endless loop tests
@@ -179,6 +265,8 @@ $vmhttpdinitsize = 0 ;
 sub chompcr
 
     {
+    local $^W = 0 ;
+
     chomp ($_[0]) ;
     if ($_[0] =~ /(.*?)\s*\r$/) 
 	{
@@ -194,18 +282,24 @@ sub chompcr
 
 sub CmpFiles 
     {
-    my ($f1, $f2) = @_ ;
+    my ($f1, $f2, $errin) = @_ ;
     my $line = 1 ;
     my $err  = 0 ;
 
     open F1, $f1 || die "***Cannot open $f1" ; 
-    open F2, $f2 || die "***Cannot open $f2" ; 
+    if (!$errin)
+	{
+	open F2, $f2 || die "***Cannot open $f2" ; 
+	}
 
     while (defined ($l1 = <F1>))
 	{
 	chompcr ($l1) ;
-	$l2 = <F2> ;
-	chompcr ($l2) ;
+	if (!$errin) 
+	    {
+	    $l2 = <F2> ;
+	    chompcr ($l2) ;
+	    }
 	if (!defined ($l2))
 	    {
 	    print "\nError in Line $line\nIs:\t$l1\nShould:\t<EOF>\n" ;
@@ -214,11 +308,18 @@ sub CmpFiles
 
 	
 	$eq = 0 ;
-	while (!$notseen && ($l2 =~ /^\^\^(.*?)$/) && !$eq)
+	while (((!$notseen && ($l2 =~ /^\^\^(.*?)$/)) || ($l2 =~ /^\^\-(.*?)$/)) && !$eq)
 	    {
 	    $l2 = $1 ;
-	    $eq = $l1 =~ /$l2/ ;
-	    $l2 = <F2> if (!$eq) ;
+	    if (($l1 =~ /^\s*$/) && ($l2 =~ /^\s*$/))
+                { 
+                $eq = 1 ;
+                }
+            else
+                {
+                $eq = $l1 =~ /$l2/ ;
+                }
+            $l2 = <F2> if (!$eq) ;
 	    chompcr ($l2) ;
 	    }
 
@@ -243,15 +344,18 @@ sub CmpFiles
 	$line++ ;
 	}
 
-    while (defined ($l2 = <F2>))
-       {
-       chompcr ($l2) ;
-       if (!($l2 =~ /^\s*$/))
-	    {
-	    print "\nError in Line $line\nIs:\t\nShould:\t$l2\n" ;
-	    return $line ;
+    if (!$errin)
+	{
+	while (defined ($l2 = <F2>))
+	   {
+	   chompcr ($l2) ;
+	   if (!($l2 =~ /^\s*$/))
+		{
+		print "\nError in Line $line\nIs:\t\nShould:\t$l2\n" ;
+		return $line ;
+		}
+	    $line++ ;
 	    }
-	$line++ ;
 	}
 
     close F1 ;
@@ -278,7 +382,11 @@ sub REQ
 	return "LWP not installed\n" ;
 	}
     
-    use HTTP::Request::Common ;
+    eval 'use HTTP::Request::Common' ;
+    if ($@)
+	{
+	return "HTTP::Request::Common not installed\n" ;
+	}
     
     
     $query ||= '' ;     
@@ -293,15 +401,21 @@ sub REQ
 	$url = new URI::URL("http://$host:$port/$loc$file?$query");
 
 	$request = new HTTP::Request($content?'POST':'GET', $url);
-
+        $request -> header ('Cookie' => $cookie) if ($cookie && !($query =~ /nocookie/)) ;
+        
 	$request -> content ($content) if ($content) ;
 	}
     else
 	{
-	$request = POST ("http://$host:$port/$loc$file?$query",
+	my @q = split (/\&|=/, $query) ;
+        
+        $request = POST ("http://$host:$port/$loc$file",
 					Content_Type => 'form-data',
-					Content      => [ upload => [undef, 'upload', Content => $upload],
-                                                          content => $content ]) ;
+					Content      => [ upload => [undef, '12upload-filename', 
+								    'Content-type' => 'test/plain',
+								    Content => $upload],
+							  content => $content,
+                                                          @q ]) ;
 	}
 	    
     #print "Request: " . $request -> as_string () ;
@@ -313,6 +427,12 @@ sub REQ
     print FH $response -> content ;
     close FH ;
 
+    my $c = $response -> header ('Set-Cookie') || '' ;
+    $cookie = $c if (!$cookie && ($c =~ /EMBPERL_UID/)) ;  
+    #print "Got Cookie $cookie\n" ;
+
+    #print $response -> headers -> as_string () ;
+
     return $response -> message if (!$response->is_success) ;
     
     return "ok" ;
@@ -361,6 +481,8 @@ sub CheckError
 	if (!($_ =~ /^\s*$/) &&
 	    !($_ =~ /\-e /) &&
 	    !($_ =~ /Warning/) &&
+	    !($_ =~ /mod_ssl\:/) &&
+	    !($_ =~ /SES\:/) &&
 	    $_ ne 'Use of uninitialized value.')
 	    {
 	    $cnt-- ;
@@ -397,23 +519,23 @@ sub CheckSVs
     seek SVLOG, -3000, 2 ;
 
     while (<SVLOG>)
-        {
-        if (/Exit-SVs: (\d+)/)
-            {
-            $num_sv = $1 || 0;
-            $last_sv[$n] ||= 0 ;
-            print "SVs=$num_sv/$last_sv[$n]/$max_sv " ;
-            if ($num_sv > $max_sv) 
-                {
-                print "GROWN " ;
-                $max_sv = $num_sv ;
+	{
+	if (/Exit-SVs: (\d+)/)
+	    {
+	    $num_sv = $1 || 0;
+	    $last_sv[$n] ||= 0 ;
+	    print "SVs=$num_sv/$last_sv[$n]/$max_sv " ;
+	    if ($num_sv > $max_sv) 
+		{
+		print "GROWN " ;
+		$max_sv = $num_sv ;
 		
-                }
-            die "\n\nMemory problem (SVs)" if ($exitonmem && $loopcnt > 2 && $last_sv[$n] != $num_sv) ;
-            $last_sv[$n] = $num_sv  ;
-            last ;
-            }
-         }
+		}
+	    die "\n\nMemory problem (SVs)" if ($exitonmem && $loopcnt > 2 && $last_sv[$n] < $num_sv) ;
+	    $last_sv[$n] = $num_sv  ;
+	    last ;
+	    }
+	 }
 
      close SVLOG ;
      }
@@ -426,29 +548,12 @@ sub CheckSVs
 #use Config qw (myconfig);
 #print myconfig () ;
 
-#### install handler which kill httpd when terminating ####
-
-BEGIN { $| = 1;
-	$SIG{__DIE__} = sub { 
-	    return unless $_[0] =~ /^\*\*\*/ ;
-	    return unless $killhttpd ;
-	    if ($EPWIN32)
-		{
-		$HttpdObj->Kill(-1) if ($HttpdObj) ;
-		}
-	    else
-		{
-		system "kill `cat $tmppath/httpd.pid` 2> /dev/null" if ($EPHTTPD ne '') ;
-		}
-	    } ;
-
-	print "\nloading...              ";
-      }
 
 ##################
 
 
 use HTML::Embperl;
+require HTML::Embperl::Module ;
 
 print "ok\n";
 
@@ -477,18 +582,22 @@ $ignoreerror = 1 if ($cmdarg =~/i/) ;
 if ($#ARGV >= 0)
     {
     if ($ARGV[0] =~ /^-/)
-        {
-        $#tests = - $ARGV[0] ;
-        }
+	{
+	$#tests = - $ARGV[0] ;
+	}
     elsif ($ARGV[0] =~ /^\d/)
-        {
-        $t = $tests[$ARGV[0]] ;
-        @tests = ($t) ;
-        }
+	{
+	@savetests = @tests ;
+	@tests = () ;
+	while ($t = shift @ARGV)
+	    {
+	    push @tests, $savetests[$t] ;
+	    }
+	}
     else
-        {
-        @tests = @ARGV ;
-        }
+	{
+	@tests = @ARGV ;
+	}
     }
     
 
@@ -520,6 +629,8 @@ $cp = HTML::Embperl::AddCompartment ('TEST') ;
 
 $cp -> deny (':base_loop') ;
 
+$ENV{EMBPERL_ALLOW} = 'asc|\\.htm$' ;
+
 do  
     {
     #############
@@ -532,13 +643,16 @@ do
 	{
 	print "\nTesting offline mode...\n\n" ;
 
-	open (SAVEERR, ">&STDERR")  || die "Cannot save stderr" ;  
-	open (STDERR, ">$offlineerr") || die "Cannot redirect stderr" ;  
-	open (ERR, "$offlineerr")  || die "Cannot open redirected stderr ($offlineerr)" ;  ;  
+	if ($loopcnt == 0)
+	    {   
+	    open (SAVEERR, ">&STDERR")  || die "Cannot save stderr" ;  
+	    open (STDERR, ">$offlineerr") || die "Cannot redirect stderr" ;  
+	    open (ERR, "$offlineerr")  || die "Cannot open redirected stderr ($offlineerr)" ;  ;  
+	    }
 
-        $n = 0 ;
-        $t_offline = 0 ;
-        $n_offline = 0 ;
+	$n = 0 ;
+	$t_offline = 0 ;
+	$n_offline = 0 ;
 	foreach $url (@tests)
 	    {
 	    ($file, $query_info, $debug, $errcnt, $option, $ns) = split (/\?/, $url) ;
@@ -548,9 +662,10 @@ do
 	    next if ($file eq 'http.htm') ;
 	    next if ($file eq 'post.htm') ;
 	    next if ($file eq 'upload.htm') ;
-	    next if ($file =~ /^exit/) ;
+	    next if ($file =~ /^exit.htm/) ;
 	    next if ($file =~ /registry/) ;
 	    next if ($file =~ /match/) ;
+	    next if ($file =~ /sess\.htm/) ;
 	    next if ($DProf && ($file =~ /safe/)) ;
 	    next if ($DProf && ($file =~ /opmask/)) ;
 
@@ -570,20 +685,19 @@ do
 			   $page, $query_info || '') ;
 	    unshift (@testargs, 'dbgbreak') if ($dbgbreak) ;
     
-	    $txt = $file . ($debug != $defaultdebug ?"-d $debug ":"") ;
-	    formline ('@<<<<<<<<<<<<<<<<<<<<... ', $txt) ;
-	    print $^A ; 
-	    $^A = '' ;
+	    $txt = $file . ($debug != $defaultdebug ?"-d $debug ":"") . '...' ;
+	    $txt .= ' ' x (30 - length ($txt)) ;
+	    print $txt ; 
     
     
 	    unlink ($outfile) ;
 
-            $n_offline++ ;
-            $t1 = HTML::Embperl::Clock () ;
+	    $n_offline++ ;
+	    $t1 = HTML::Embperl::Clock () ;
 	    $err = HTML::Embperl::run (@testargs) ;
 	    $t_offline += HTML::Embperl::Clock () - $t1 ;
 
-            if ($memcheck)
+	    if ($memcheck)
 		{
 		my $vmsize = GetMem ($$) ;
 		$vminitsize = $vmsize if $loopcnt == 2 ;
@@ -591,22 +705,25 @@ do
 		print "GROWN! at iteration = $loopcnt  " if ($vmsize > $vmmaxsize) ;
 		$vmmaxsize = $vmsize if ($vmsize > $vmmaxsize) ;
 		CheckSVs ($loopcnt, $n) ;
-                }
+		}
 		
-	    $err = CheckError ($errcnt) if ($err == 0) ;
+	    $errin = $err ;
+	    $err = CheckError ($errcnt) if ($err == 0 || ($errcnt > 0 && $err == 500) || $file eq 'notfound.htm'  || $file eq 'notallow.xhtm') ;
     
-	    if ($err == 0)
+	    
+	    if ($err == 0 && $errin != 500 && $file ne 'notfound.htm' && $file ne 'notallow.xhtm')
 		{
 		$page =~ /.*\/(.*)$/ ;
 		$org = "$cmppath/$1" ;
 
-		$err = CmpFiles ($outfile, $org) ;
+		$err = CmpFiles ($outfile, $org, $errin) ;
 		}
 
 	    print "ok\n" unless ($err) ;
+	    $err = 0 if ($ignoreerror) ;
 	    last if $err ;
 	    $n++ ;
-            }
+	    }
 	}
     
     if ($testtype =~ /e/)
@@ -639,12 +756,12 @@ do
 		}
 
 
-	    formline ('@<<<<<<<<<<<<<<<<<<<<... ', "$txt from file") ;
-	    print $^A ; 
-	    $^A = '' ;
+	    $txt2 = "$txt from file...";
+	    $txt2 .= ' ' x (30 - length ($txt2)) ;
+	    print $txt2 ; 
 
 	    unlink ($outfile) ;
-            $t1 = HTML::Embperl::Clock () ;
+	    $t1 = HTML::Embperl::Clock () ;
 	    $err = HTML::Embperl::Execute ({'inputfile'  => $src,
 					    'mtime'      => 1,
 					    'outputfile' => $outfile,
@@ -659,19 +776,19 @@ do
 
 	    if ($err == 0)
 		{
-		formline ('@<<<<<<<<<<<<<<<<<<<<... ', "$txt from memory") ;
-		print $^A ;     
-		$^A = '' ;
+		$txt2 = "$txt from memory...";
+		$txt2 .= ' ' x (30 - length ($txt2)) ;
+		print $txt2 ; 
 
 		unlink ($outfile) ;
-                $t1 = HTML::Embperl::Clock () ;
+		$t1 = HTML::Embperl::Clock () ;
 		$err = HTML::Embperl::Execute ({'input'      => \$indata,
 						'inputfile'  => 'i1',
 						'mtime'      => 1,
 						'outputfile' => $outfile,
 						'debug'      => $defaultdebug,
 						}) ;
-	        $t_exec += HTML::Embperl::Clock () - $t1 ; 
+		$t_exec += HTML::Embperl::Clock () - $t1 ; 
 		    
 		$err = CheckError ($errcnt) if ($err == 0) ;
 		$err = CmpFiles ($outfile, $org)  if ($err == 0) ;
@@ -680,19 +797,20 @@ do
 
 	    if ($err == 0)
 		{
-		formline ('@<<<<<<<<<<<<<<<<<<<<... ', "$txt to memory") ;
-		print $^A ;     
-		$^A = '' ;
+		$txt2 = "$txt to memory...";
+		$txt2 .= ' ' x (30 - length ($txt2)) ;
+		print $txt2 ; 
 
 		my $outdata ;
+                my @errors ;
 		unlink ($outfile) ;
-                $t1 = HTML::Embperl::Clock () ;
+		$t1 = HTML::Embperl::Clock () ;
 		$err = HTML::Embperl::Execute ({'inputfile'  => $src,
 						'mtime'      => 1,
 						'output'     => \$outdata,
 						'debug'      => $defaultdebug,
 						}) ;
-	        $t_exec += HTML::Embperl::Clock () - $t1 ; 
+		$t_exec += HTML::Embperl::Clock () - $t1 ; 
 		    
 		$err = CheckError ($errcnt) if ($err == 0) ;
 	
@@ -705,23 +823,30 @@ do
 
 	    if ($err == 0)
 		{
-		formline ('@<<<<<<<<<<<<<<<<<<<<... ', "$txt from/to memory") ;
-		print $^A ;     
-		$^A = '' ;
+		$txt2 = "$txt from/to memory...";
+		$txt2 .= ' ' x (30 - length ($txt2)) ;
+		print $txt2 ; 
 
 		my $outdata ;
 		unlink ($outfile) ;
-                $t1 = HTML::Embperl::Clock () ;
+		$t1 = HTML::Embperl::Clock () ;
 		$err = HTML::Embperl::Execute ({'input'      => \$indata,
 						'inputfile'  => $src,
 						'mtime'      => 1,
 						'output'     => \$outdata,
+		                                'errors'     => \@errors,
 						'debug'      => $defaultdebug,
 						}) ;
-	        $t_exec += HTML::Embperl::Clock () - $t1 ; 
+		$t_exec += HTML::Embperl::Clock () - $t1 ; 
 		    
 		$err = CheckError ($errcnt) if ($err == 0) ;
 	
+                if (@errors != 0)
+                    {
+                    print "\n\n\@errors does not return correct number of errors (is " . scalar(@errors) . ", should 0)\n" ;
+                    $err = 1 ;
+                    }
+
 		open FH, ">$outfile" or die "Cannot open $outfile ($!)" ;
 		print FH $outdata ;
 		close FH ;
@@ -739,22 +864,30 @@ do
 
 	    if ($err == 0)
 		{
-		formline ('@<<<<<<<<<<<<<<<<<<<<... ', "$txt to memory") ;
-		print $^A ;     
-		$^A = '' ;
+		$txt2 = "$txt to memory...";
+		$txt2 .= ' ' x (30 - length ($txt2)) ;
+		print $txt2 ; 
 
 		my $outdata ;
+                my @errors ;
 		unlink ($outfile) ;
-                $t1 = HTML::Embperl::Clock () ;
+		$t1 = HTML::Embperl::Clock () ;
 		$err = HTML::Embperl::Execute ({'inputfile'  => $src,
 						'mtime'      => 1,
 						'output'     => \$outdata,
 						'debug'      => $defaultdebug,
-						}) ;
-	        $t_exec += HTML::Embperl::Clock () - $t1 ; 
+		                                'errors'     => \@errors,
+                				}) ;
+		$t_exec += HTML::Embperl::Clock () - $t1 ; 
 		    
-		$err = CheckError (7) if ($err == 0) ;
-	
+                $err = CheckError (8) if ($err == 0) ;
+
+                if (@errors != 12)
+                    {
+                    print "\n\n\@errors does not return correct number of errors (is " . scalar(@errors) . ", should 12)\n" ;
+                    $err = 1 ;
+                    }
+
 		open FH, ">$outfile" or die "Cannot open $outfile ($!)" ;
 		print FH $outdata ;
 		close FH ;
@@ -766,7 +899,7 @@ do
 	    }
 	}
 
-    if (($testtype =~ /e/) || ($testtype =~ /o/))
+    if ((($testtype =~ /e/) || ($testtype =~ /o/)) && $looptest == 0)
 	{
 	close STDERR ;
 	open (STDERR, ">&SAVEERR") ;
@@ -800,8 +933,8 @@ do
 	$cf = <IFH> ;
 	close IFH ;
 	open OFH, ">$httpdconf" or die "***Cannot open $httpconf" ;
-	eval "print OFH \"$cf\"" ;
-	die "***Cannot print $httpconf ($@)" if ($@) ;
+	eval $cf ;
+	die "***Cannot eval $httpconf ($@)" if ($@) ;
 	close OFH ;
 	$/ = $rs ;
     
@@ -815,9 +948,10 @@ do
 	if ($EPWIN32)
 	    {
 	    $ENV{PATH} .= ";$EPHTTPDDLL" if ($EPWIN32) ;
+	    $ENV{PERL_STARTUP_DONE} = 1 ;
 
 	    Win32::Process::Create($HttpdObj, $EPHTTPD,
-				   "Apache  $XX -f $EPPATH/$httpdconf ", 0,
+				   "Apache -s $XX -f $EPPATH/$httpdconf ", 0,
 				   # NORMAL_PRIORITY_CLASS,
 				   0,
 				    ".") or die "***Cannot start $EPHTTPD" ;
@@ -833,8 +967,15 @@ do
 	    if (!open FH, "$tmppath/httpd.pid")
 		{
 		sleep (7) ;
-		open FH, "$tmppath/httpd.pid" or die "Cannot open $tmppath/httpd.pid" ;
-		}
+		if (!open FH, "$tmppath/httpd.pid")
+                    {
+            	    open (FERR, "$httpderr") ;  
+                    print $_ while (<FERR>) ;
+                    close FERR ;
+                    die "Cannot open $tmppath/httpd.pid" ;
+		    }
+                }
+
 	    }
 	$httpdpid = <FH> ;
 	chop($httpdpid) ;       
@@ -860,9 +1001,10 @@ do
 	else
 	    { print "\nTesting cgi mode...\n\n" ; }
 
+	$cookie = undef ;
         $t_req = 0 ;
-        $n_req = 0 ;
-        $n = 0 ;
+	$n_req = 0 ;
+	$n = 0 ;
 	foreach $url (@tests)
 	    {
 	    ($file, $query_info, $debug, $errcnt) = split (/\?/, $url) ;
@@ -870,15 +1012,29 @@ do
 	    next if ($file =~ /\// && $loc eq $cgiloc) ;        
 	    next if ($file eq 'taint.htm' && $loc eq $cgiloc) ;
 	    next if ($file eq 'reqrec.htm' && $loc eq $cgiloc) ;
-	    next if (($file =~ /^exit/) && $loc eq $cgiloc) ;
+	    next if (($file =~ /^exit.htm/) && $loc eq $cgiloc) ;
 	    #next if ($file eq 'error.htm' && $loc eq $cgiloc && $errcnt < 16) ;
 	    next if ($file eq 'varerr.htm' && $loc eq $cgiloc && $errcnt > 0) ;
 	    next if ($file eq 'varerr.htm' && $looptest) ;
 	    next if (($file =~ /registry/) && $loc eq $cgiloc) ;
 	    next if (($file =~ /match/) && $loc eq $cgiloc) ;
-	    next if ($file eq 'http.htm' && $loc eq $cgiloc) ;
+	    #next if ($file eq 'http.htm' && $loc eq $cgiloc) ;
 	    next if ($file eq 'chdir.htm' && $EPWIN32) ;
+	    next if ($file eq 'notfound.htm' && $loc eq $cgiloc && $EPWIN32) ;
+	    #next if ($file eq 'notallow.xhtm' && $loc eq $cgiloc && $EPWIN32) ;
 	    next if ($file =~ /opmask/ && $EPSTARTUP =~ /_dso/) ;
+	    next if ($file eq 'clearsess.htm' && !$looptest) ;
+	    if ($file =~ /sess\.htm/)
+                { 
+                next if ($loc eq $cgiloc && $EPSESSIONCLASS ne 'Embperl') ;
+                if (!$EPSESSIONVERSION)
+                    {
+		    $txt2 = "$file...";
+		    $txt2 .= ' ' x (29 - length ($txt2)) ;
+		    print "$txt2 skip on this plattform\n" ; 
+                    next ;
+                    }
+                }
      
 	    $debug ||= $defaultdebug ;  
 	    $errcnt ||= 0 ;
@@ -899,25 +1055,24 @@ do
 		$notseen = 1 ;
 		}
     
-	    $txt = "$file" . ($debug != $defaultdebug ?"-d $debug ":"") ;
-	    formline ('@<<<<<<<<<<<<<<<<<<<<... ', $txt) ;
-	    print $^A ; 
-	    $^A = '' ;
+	    $txt = "$file" . ($debug != $defaultdebug ?"-d $debug ":"") . '...' ;
+	    $txt .= ' ' x (30 - length ($txt)) ;
+	    print $txt ; 
 	    unlink ($outfile) ;
 	    
 	    $content = undef ;
 	    $content = "f1=abc1&f2=1234567890&f3=" . 'X' x 8192 if ($file eq 'post.htm') ;
 	    $upload = undef ;
 	    if ($file eq 'upload.htm') 
-                {
-                $upload = "f1=abc1\r\n&f2=1234567890&f3=" . 'X' x 8192 ;
-                $content = "Hi there!" ;
-                }
+		{
+		$upload = "f1=abc1\r\n&f2=1234567890&f3=" . 'X' x 8192 ;
+		$content = "Hi there!" ;
+		}
 
-            $n_req++ ;
-            $t1 = HTML::Embperl::Clock () ;
+	    $n_req++ ;
+	    $t1 = HTML::Embperl::Clock () ;
 	    $m = REQ ($loc, $file, $query_info, $outfile, $content, $upload) ;
-            $t_req += HTML::Embperl::Clock () - $t1 ; 
+	    $t_req += HTML::Embperl::Clock () - $t1 ; 
 
 	    if ($memcheck)
 		{
@@ -929,7 +1084,7 @@ do
 		$vmhttpdsize = $vmsize if ($vmsize > $vmhttpdsize) ;
 		CheckSVs ($loopcnt, $n) ;
 		
-                }
+		}
 	    if (($m || '') ne 'ok' && $errcnt == 0)
 		{
 		$err = 1 ;
@@ -937,8 +1092,9 @@ do
 		last ;
 		}
 
-	    $err = CheckError ($errcnt) if ($err == 0 && $checkerr) ;
-	    if ($err == 0)
+	    #$errcnt++ if ($loc eq $cgiloc && $file eq 'notallow.xhtm') ;   
+	    $err = CheckError ($errcnt) if (($err == 0 || $file eq 'notfound.htm' || $file eq 'notallow.xhtm') && $checkerr ) ;
+	    if ($err == 0 && $file ne 'notfound.htm' && $file ne 'notallow.xhtm')
 		{
 		$page =~ /.*\/(.*)$/ ;
 		$org = "$cmppath/$1" ;
@@ -949,38 +1105,38 @@ do
 
 	    print "ok\n" unless ($err) ;
 	    $err = 0 if ($ignoreerror) ;
-            last if ($err) ;
+	    last if ($err) ;
 	    $n++ ;
-            }
+	    }
 
 	if ($loc ne $cgiloc)   
 	    { 
-            $t_mp = $t_req ;
-            $n_mp = $n_req ;
-            }
+	    $t_mp = $t_req ;
+	    $n_mp = $n_req ;
+	    }
 	else
 	    {
-            $t_cgi = $t_req ;
-            $n_cgi = $n_req ;
-            }
+	    $t_cgi = $t_req ;
+	    $n_cgi = $n_req ;
+	    }
 
 	if ($testtype =~ /c/ && $err == 0 && $loc ne $cgiloc && $loopcnt == 0)   
 	    { 
-            $loc = $cgiloc ;
-            }
+	    $loc = $cgiloc ;
+	    }
 	else
 	    {
-            $loc = '' ;
-            }
+	    $loc = '' ;
+	    }
 	}
 
     if ($defaultdebug == 0)
-        {
-        print "\n" ;
-        print "Offline:  $n_offline tests takes $t_offline sec = ", int($t_offline / $n_offline * 1000) / 1000.0, " sec per test\n" if ($t_offline) ;
-        print "mod_perl: $n_mp tests takes $t_mp sec = ", int($t_mp / $n_mp * 1000) / 1000.0 , " sec per test\n"  if ($t_mp) ;
-        print "CGI:      $n_cgi tests takes $t_cgi sec = ", int($t_cgi / $n_cgi * 1000) / 1000.0 , " sec per test\n"  if ($t_cgi) ;
-        }
+	{
+	print "\n" ;
+	print "Offline:  $n_offline tests takes $t_offline sec = ", int($t_offline / $n_offline * 1000) / 1000.0, " sec per test\n" if ($t_offline) ;
+	print "mod_perl: $n_mp tests takes $t_mp sec = ", int($t_mp / $n_mp * 1000) / 1000.0 , " sec per test\n"  if ($t_mp) ;
+	print "CGI:      $n_cgi tests takes $t_cgi sec = ", int($t_cgi / $n_cgi * 1000) / 1000.0 , " sec per test\n"  if ($t_cgi) ;
+	}
 
     $loopcnt++ ;
     }
